# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""Implementation of Opus."""
# based off the implementation of discord.py and disco.
import array
import ctypes
import ctypes.util
import os
import struct
import sys
import threading
from enum import Enum
from typing import List

from ..internal.exceptions import DiscordError

__all__: List[str] = ["Encoder"]

# structs
class EncoderStruct(ctypes.Structure):  # noqa: ignore

    pass


# if your curious, the 'decoder' handles, voice recv.
class DecoderStuct(ctypes.Structure):  # noqa: ignore

    pass


# pointers
c_int_ptr = ctypes.POINTER(ctypes.c_int)
c_int16_ptr = ctypes.POINTER(ctypes.c_int16)
c_float_ptr = ctypes.POINTER(ctypes.c_float)

EncoderStructPtr = ctypes.POINTER(EncoderStruct)
DecoderStructPtr = ctypes.POINTER(DecoderStuct)

opus_loaded: threading.Event = threading.Event()


# options
default_options = {
    # generic stuff
    "opus_get_version_string": (None, ctypes.c_char_p, None),
    "opus_strerror": ([ctypes.c_int], ctypes.c_char_p, None),
    # packets
    "opus_packet_get_bandwidth": ([ctypes.c_char_p], ctypes.c_int),
    "opus_packet_get_nb_channels": ([ctypes.c_char_p], ctypes.c_int),
    "opus_packet_get_nb_frames": ([ctypes.c_char_p, ctypes.c_int], ctypes.c_int),
    "opus_packet_get_samples_per_frame": (
        [ctypes.c_char_p, ctypes.c_int],
        ctypes.c_int,
    ),
}

# Enums


class AppEnum(Enum):
    AUDIO = 2049
    VOIP = 2048
    LOWDELAY = 2051


class CntrlEnum(Enum):
    SET_BITRATE = 4002
    SET_BANDWIDTH = 4008
    SET_FEC = 4012
    SET_PLP = 4014


# Errors


class OpusError(DiscordError):
    pass


# loader


def loader(name: str):
    opus = ctypes.cdll.LoadLibrary(name)

    return opus


def load_opus() -> bool:
    global lib
    try:
        if sys.platform == "win32":
            basedir = os.path.dirname(os.path.abspath(__file__))
            bitness = struct.calcsize("P") * 8
            target = "x64" if bitness > 32 else "x86"
            filename = os.path.join(basedir, "bin", f"opus-{target}.dll")
            lib = loader(filename)
            opus_loaded.set()
        else:
            lib = loader(ctypes.util.find_library("opus"))
            opus_loaded.set()
    except Exception as exc:
        lib = None
        raise OpusError from exc

    return lib is not None


def check_load():
    if opus_loaded.is_set() is not True:
        return load_opus()
    else:
        return lib


class Opus:
    EXP = {}

    def __init__(self):
        self.lib = check_load()
        meths = {}
        meths.update(default_options)
        meths.update(self.EXP)

        for name, option in meths.items():  # type: ignore # noqa: ignore
            func = getattr(loader, name)

            try:
                if option[1]:
                    func.argtypes = option[0]

                func.restype = option[1]

                setattr(self, name, func)
            except KeyError:
                pass


# encoder
class Encoder(Opus):
    """A class used to encode `.pcm` data into opus data for the discord voice gateway"""

    EXP = {
        "opus_encoder_get_size": ([ctypes.c_int], ctypes.c_int, None),
        "opus_encoder_create": (
            [ctypes.c_int, ctypes.c_int, ctypes.c_int, c_int_ptr],
            EncoderStructPtr,
        ),
        "opus_encode": (
            [
                EncoderStructPtr,
                c_int16_ptr,
                ctypes.c_int,
                ctypes.c_char_p,
                ctypes.c_int32,
            ],
            ctypes.c_int32,
        ),
        "opus_encode_float": (
            [
                EncoderStructPtr,
                c_float_ptr,
                ctypes.c_int,
                ctypes.c_char_p,
                ctypes.c_int32,
            ],
            ctypes.c_int32,
        ),
        "opus_encoder_ctl": (None, ctypes.c_int32),
        "opus_encoder_destroy": ([EncoderStructPtr], None, None),
    }

    def __init__(self):
        self._inst = None

    @property
    def inst(self):
        if not self._inst:
            self._inst = self.create()
            self.set_bitrate(128)
            self.set_fec(True)
            self.set_expected_plp(0.15)

        return self._inst

    def set_bitrate(self, kbps):
        kbps = min(128, max(16, int(kbps)))
        ret = self.opus_encoder_ctl(
            self.inst, int(CntrlEnum.SET_BITRATE.value), kbps * 1024
        )

        if ret < 0:
            raise RuntimeError("Failed to set bitrate, %s: %s", kbps, ret)

    def set_fec(self, value):
        ret = self.opus_encoder_ctl(self.inst, int(CntrlEnum.SET_FEC.value), int(value))

        if ret < 0:
            raise RuntimeError("Failed to set bitrate, %s: %s", value, ret)

    def set_expected_plp(self, perc):
        ret = self.opus_encoder_ctl(
            self.inst, int(CntrlEnum.SET_PLP.value), min(100, max(0, int(perc * 100)))
        )

        if ret < 0:
            raise RuntimeError("Failed to set bitrate, %s: %s", perc, ret)

    def create(self):
        ret = ctypes.c_int()
        result = self.opus_encoder_create(
            self.sampling_rate, self.channels, self.application.value, ctypes.byref(ret)
        )

        if ret.value != 0:
            raise RuntimeError

        return result

    def __del__(self):
        if hasattr(self, "_inst") and self._inst:
            self.opus_encoder_destroy(self._inst)
            self._inst = None

    def encode(self, pcm, frame_size):
        m_data_bytes = len(pcm)
        pcm = ctypes.cast(pcm, c_int16_ptr)
        data = (ctypes.c_char * m_data_bytes)()

        ret = self.opus_encode(self.inst, pcm, frame_size, data, m_data_bytes)

        if ret < 0:
            raise RuntimeError

        return array.array("b", data[:ret].tobytes())



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

import asyncio
from typing import Any, Callable, Coroutine, List, Optional

from discord.types.dict import Dict

from ..internal.dispatcher import Dispatcher
from ..state import ConnectionState
from .interaction import Interaction

__all__: List[str] = ["ApplicationCommandRegistry"]


class ApplicationCommandRegistry:
    """Handles registration of application commands."""

    def __init__(self, rest_factory, state: ConnectionState):
        self.factory = rest_factory
        self.state = state
        self.dispatcher = Dispatcher(state=self.state)
        self.state.loop.create_task(self.on_ready())
        self.unregistered_commands: asyncio.Event = asyncio.Event()

    async def run(
        self, coro: Callable[..., Coroutine[Any, Any, Any]], data, state, cog=None
    ) -> None:
        try:
            if cog:
                i = Interaction(data, state)
                await coro(cog, i)
            else:
                i = Interaction(data, state)
                await coro(i)
        except asyncio.CancelledError:
            pass
        except Exception:
            raise

    async def on_ready(self):
        await asyncio.sleep(20)
        glob = await self.factory.get_global_application_commands(self.state._bot_id)

        await self.check_application_commands(glob)

        await asyncio.sleep(10)

        for guild in self.state.guilds._cache.values():
            commands = await self.factory.get_guild_application_commands(
                self.state._bot_id, guild["id"]
            )
            for command in commands:
                if command not in self.state.application_commands.items():
                    await self.factory.delete_guild_application_command(
                        self.state._bot_id, guild["id"], command["id"]
                    )
        self.unregistered_commands.set()

    async def check_application_commands(self, rglobal):
        for command in rglobal:
            await self.factory.delete_global_application_command(
                self.state._bot_id, command["id"]
            )

    async def register_guild_slash_command(
        self,
        guild_id,
        name,
        description,
        callback,
        cog: Any = None,
        options: Optional[List[Dict]] = None,
        default_permission: bool = True,
    ) -> dict:
        if not self.unregistered_commands.is_set():
            await self.unregistered_commands.wait()
        r = await self.factory.create_guild_application_command(
            self.state._bot_id,
            guild_id,
            name,
            description,
            options,
            default_permission,
            type=1,
        )
        self.state.application_commands[name] = {
            "d": r,
            "callback": callback,
            "self": self,
            "cog": cog,
        }
        return r

    async def register_global_slash_command(
        self,
        name,
        description,
        callback,
        cog: Any = None,
        options: Optional[List[Dict]] = None,
        default_permission: bool = True,
    ) -> dict:
        if not self.unregistered_commands.is_set():
            await self.unregistered_commands.wait()
        r = await self.factory.create_global_application_command(
            self.state._bot_id, name, description, options, default_permission, type=1
        )
        self.state.application_commands[name] = {
            "d": r,
            "callback": callback,
            "self": self,
            "cog": cog,
        }
        return r



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""The Voice Client."""
import asyncio
import logging
import socket
import struct
import threading
from typing import List, Optional, Tuple

from discord import utils

from ..channels import VoiceChannel
from ..state import ConnectionState
from .gateway import VoiceGateway
from .opus import Encoder
from .players import AudioPlayer

has_nacl: bool

try:
    from nacl import secret as nacl  # noqa: ignore
except (ImportError, ModuleNotFoundError):
    has_nacl = False
else:
    has_nacl = True

_log = logging.getLogger(__name__)


class VoiceClient:
    """Used to interact with the discord voice api

    Parameters
    ----------
    state
        The :class:`ConnectionState`
    dispatcher
        The :class:`Dispatcher`
    gateway
        The :class:`Gateway`
    """

    endpoint_ip: str
    voice_port: int
    secret_key: List[int]
    ssrc: int

    def __init__(self, state: ConnectionState, channel: VoiceChannel):
        if not has_nacl:
            raise RuntimeError("PyNaCl is needed in order to use voice")

        self.token: str = None
        self.socket = None
        self.loop: asyncio.AbstractEventLoop = state.loop
        self._state: ConnectionState = state
        self._connected: threading.Event = threading.Event()

        self._handshaking: bool = False
        self._potentially_reconnecting: bool = False
        self._voice_state_complete: asyncio.Event = asyncio.Event()
        self._voice_server_complete: asyncio.Event = asyncio.Event()

        self.mode: str = None
        self._connections: int = 0
        self.sequence: int = 0
        self.timestamp: int = 0
        self.channel = channel
        self.timeout: float = 0
        self._runner: asyncio.Task = None
        self._player: Optional[AudioPlayer] = None
        self.encoder: Encoder = None
        self._lite_nonce: int = 0
        self.ws: VoiceGateway = None
        self._state.app.dispatcher.add_listener(
            self.on_voice_state_update, "on_raw_voice_state_update"
        )
        self._state.app.dispatcher.add_listener(
            self.on_voice_server_update, "on_raw_voice_server_update"
        )

    warn_nacl = not has_nacl
    supported_modes: Tuple[str] = (
        "xsalsa20_poly1305_lite",
        "xsalsa20_poly1305_suffix",
        "xsalsa20_poly1305",
    )

    async def on_voice_state_update(self, data):
        self.session_id = data["session_id"]
        channel_id = data["channel_id"]

        if not self._handshaking or self._potentially_reconnecting:
            if channel_id is None:
                await self.disconnect()
            else:
                guild = self.guild_id
                self.channel = (
                    channel_id and guild and guild._factory.get_channel(int(channel_id))
                )
        else:
            self._voice_state_complete.set()

    async def on_voice_server_update(self, data: dict):
        if self._voice_server_complete.is_set():
            return

        self.token = data.get("token")
        self.guild_id = int(data["guild_id"])
        endpoint: str = data.get("endpoint")

        if endpoint is None or self.token is None:
            return

        self.endpoint, _, _ = endpoint.rpartition(":")
        if self.endpoint.startswith("wss://"):
            self.endpoint = self.endpoint[6:]

        self.endpoint_ip = None

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setblocking(False)

        if not self._handshaking:
            await self.ws.close(4000)
            return

        self._voice_server_complete.set()

    async def voice_connect(self):
        await self.channel.guild.change_voice_state(channel=self.channel.id)

    async def voice_disconnect(self):
        await self.channel.guild.change_voice_state(channel=None)

    def prepare_handshake(self):
        self._voice_server_complete.clear()
        self._voice_state_complete.clear()
        self._handshaking = True
        self._connections += 1

    def finish_handshake(self):
        _log.info("Hand shaking done, using endpoint: %s", self.endpoint)
        self._handshaking = False
        self._voice_server_complete.clear()
        self._voice_state_complete.clear()

    async def ws_connect(self) -> VoiceGateway:
        ws = await VoiceGateway.voice_client_entry(self)
        self._connected.clear()
        await ws.connect()
        self._connected.set()
        return ws

    async def connect(self, *, reconnect: bool, timeout: float):
        self.timeout = timeout

        for i in range(5):
            self.prepare_handshake()

            futures = [
                self._voice_server_complete.wait(),
                self._voice_state_complete.wait(),
            ]

            await self.voice_connect()

            try:
                await utils.wait_for(futures, timeout=timeout)
            except asyncio.TimeoutError:
                await self.disconnect(force=True)
                raise

            self.finish_handshake()

            try:
                self.ws = await self.ws_connect()
                break
            except asyncio.TimeoutError:
                if reconnect:
                    _log.error("Failed to connect to the voice gateway, retrying...")
                    await asyncio.sleep(1 + i * 2)
                    await self.voice_disconnect()
                    continue
                else:
                    raise

    async def disconnect(self, *, force: bool = False):
        if not force and not self.is_connected():
            return

        self._connected.clear()

        try:
            if self.ws:
                await self.ws.close(1000)

            await self.voice_disconnect()
        finally:
            if self.socket:
                self.socket.close()

    async def move_to(self, channel: int):
        await self.channel.guild.change_voice_state(channel=channel)

    def is_connected(self) -> bool:
        return self._connected.is_set()

    # encryption things

    def checked_add(self, attr, value, limit):
        val = getattr(self, attr)
        if val + value > limit:
            setattr(self, attr, 0)
        else:
            setattr(self, attr, val + value)

    def _get_voice_packet(self, data):
        byteheader = bytearray(12)

        byteheader[0] = 0x80
        byteheader[1] = 0x78
        struct.pack_into(">H", byteheader, 2, self.sequence)
        struct.pack_into(">I", byteheader, 4, self.timestamp)
        struct.pack_into(">I", byteheader, 8, self.ssrc)

        encrypt_packet = getattr(self, "__encrypt__" + self.mode)
        return encrypt_packet(byteheader, data)

    def send_audio_packet(self, data: bytes, *, encode: bool = True):
        self.checked_add("sequence", 1, 65535)
        if encode:
            encoded_data = self.encoder.encode(data, 960)
        else:
            encoded_data = data

        packet = self._get_voice_packet(data)
        try:
            self.socket.sendto(packet, (self.endpoint_ip, self.voice_port))
        except BlockingIOError:
            _log.error("A packet was dropped")

        self.checked_add("timestamp", 65535, 4294967295)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
# Implementation of sending Files via rest.
import io
import os
import typing as t

from .internal.exceptions import DiscordError, Forbidden

__all__: t.List[str] = ["File"]


class File:
    """Represents a Discord file.

    .. versionadded:: 0.4.0

    Parameters
    ----------
    fp
        The File path, can be :class:`str`, :class:`bytes`,
        :class:`os.PathLike`, :class:`os.BufferedIOBase`.
    filename
        The filename, defaults `None`,
    spoiler
        If the file should be seeable or not.
    """

    def __init__(
        self,
        fp: t.Union[str, bytes, os.PathLike, io.BufferedIOBase],
        *,
        filename: t.Optional[str] = None,
        spoiler: bool = False,
    ):
        if isinstance(fp, io.IOBase):
            if not (fp.seekable() and fp.readable()):
                raise DiscordError(
                    f"File buffer {fp!r} must be both seekable and readable."
                )
            self.fp = fp
            self._og_pos = fp.tell()
            self._owner = False
        else:
            self.fp = open(fp, "rb")
            self._og_pos = 0
            self._owner = True

        self._closer = self.fp.close()
        self.fp.close = lambda: None

        if filename is None:
            if isinstance(fp, str):
                _, self.filename = os.path.split(fp)
            else:
                self.filename = getattr(fp, "name", None)
        else:
            self.filename = filename

        if (
            spoiler
            and self.filename is not None  # noqa: ignore
            and not self.filename.startswith("SPOILER_")  # noqa: ignore
        ):
            self.filename = "SPOILER_" + self.filename

        self.spoiler = spoiler or (
            self.filename is not None and self.filename.startswith("SPOILER_")
        )

    def reset(self, *, seek: t.Union[int, bool] = True) -> None:
        if seek:
            self.fp.seek(self._og_pos)

    def close(self) -> None:
        self.fp.close = self._closer
        if self._owner:
            self._closer()
        else:
            raise Forbidden("You arent allowed to close this file")



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

from ..assets import Emoji, Sticker
from ..guild import Guild, Role, ScheduledEvent
from ..member import Member
from ..user import User
from .core import Event


# join/leave/update
class OnGuildJoin(Event):
    """A event that gets called when you join a guild

    Returns
    -------
    Guild: :class:`Guild`
    """

    def process(self):
        ret = Guild(self.data, self.state.app.factory)

        self.dispatch("GUILD_JOIN", ret)


class OnGuildUpdate(Event):
    """A event which is called when a guild is updated

    Returns
    -------
    before: :class:`Guild`
    after: :class:`Guild`
    """

    def process(self):
        before = Guild(self.state.guilds.get(self.data["id"]), self.state.app.factory)
        after = Guild(self.data, self.state.app.factory)
        self.state.guilds.edit(self.data["id"], self.data)

        self.dispatch("GUILD_UPDATE", before, after)


class OnGuildLeave(Event):
    """Gets called when your bot leaves or gets kicked from a guild

    Returns
    -------
    Guild: :class:`Guild`
    """

    def process(self):
        guild = Guild(self.state.guilds.pop(self.data["id"]), self.state.app.factory)
        self.dispatch("GUILD_LEAVE", guild)


# bans
class OnGuildBan(Event):
    """Gets called when a member is banned from a guild

    Returns
    -------
    User: :class:`User`
    Guild: :class:`Guild`
    """

    def process(self):
        user = User(self.data["user"])
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )

        self.dispatch("GUILD_BAN", user, guild)


class OnGuildBanRemove(Event):
    """Gets called when a guild ban gets removed

    Returns
    -------
    User: :class:`User`
    Guild: :class:`Guild`
    """

    def process(self):
        user = User(self.data["user"])
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )

        self.dispatch("GUILD_BAN_REMOVE", user, guild)


class OnGuildIntegrationsUpdate(Event):
    """Called when a integration is updated

    Returns
    -------
    Guild: :class:`Guild`
    """

    def process(self):
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )

        self.dispatch("GUILD_INTEGRATIONS_UPDATE", guild)


# assets


class OnGuildEmojisUpdate(Event):
    """Called when 1 or more guild emojis is updated

    Returns
    -------
    Emojis: list[:class:`Emoji`]
    Guild: :class:`Guild`
    """

    def process(self):
        emojis = [Emoji(emoji) for emoji in self.data["emojis"]]
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )
        self.dispatch("GUILD_EMOJIS_UPDATE", emojis, guild)


class OnGuildStickersUpdate(Event):
    """Called when 1 or more guild stickers are updated

    Returns
    -------
    Stickers: list[:class:`Sticker`]
    Guild: :class:`Guild`
    """

    def process(self):
        stickers = [Sticker(sticker, self.state) for sticker in self.data["stickers"]]
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )

        self.dispatch("GUILD_STICKERS_UPDATE", stickers, guild)


# members
class OnMemberJoin(Event):
    """Gets called when a member joins a guild

    Returns
    -------
    Member: :class:`Member`
    Guild: :class:`Guild`
    """

    def process(self):
        # it says inner payload so i would guess it's a member keyword?
        member = Member(self.data, self.data["guild_id"], self.state.app.factory)
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )

        self.dispatch("MEMBER_JOIN", member, guild)


class OnMemberLeave(Event):
    """Gets called when a member leaves a guild

    Returns
    -------
    User: :class:`User`
    Guild: :class:`Guild`
    """

    def process(self):
        user = User(self.data["user"])
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )

        for member in self.state.members._cache.values():
            if member.user.id == user.id:
                self.state.members.pop(member)

        self.dispatch("MEMBER_LEAVE", user, guild)


class OnMemberUpdate(Event):
    """Called when a member is updated

    Returns
    -------
    before: :class:`Member`, could be :class:`None`
    after: :class:`Member`
    """

    def process(self):
        before = None
        for member in self.state.members._cache.values():
            if member.user.id == self.data["user"]["id"]:
                before = member
                self.state.members.pop(member)

        after = Member(self.data, self.data["guild_id"], self.state.app.factory)

        self.dispatch("MEMBER_UPDATE", before, after)


# roles
class OnRoleCreate(Event):
    """Gets called when a role is created

    Returns
    -------
    Role: :class:`Role`
    Guild: :class:`Guild`
    """

    def process(self):
        role = Role(self.data["role"])
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )

        self.state.roles.new(role.id, role)

        self.dispatch("ROLE_CREATE", role, guild)


class OnRoleUpdate(Event):
    """Gets called when a role is updated

    Returns
    -------
    before: :class:`Role`, can be :class:`None`
    after: :class:`Role`
    Guild: :class:`Guild`
    """

    def process(self):
        before = Role(self.state.roles.get(self.data["role"]["id"]))
        after = Role(self.data["role"])
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )
        self.state.roles.edit(after.id, after)

        self.dispatch("ROLE_UPDATE", before, after, guild)


class OnRoleDelete(Event):
    """Gets caleld when a role is deleted

    Returns
    -------
    role: :class:`Role`, can be :class:`None`
    Guild: :class:`Guild`
    """

    def process(self):
        role = Role(self.state.roles.pop(self.data["role_id"]))
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )

        self.dispatch("ROLE_DELETE", role, guild)


class OnScheduledEventCreate(Event):
    """Gets called when a scheduled event is created

    Returns
    -------
    Event: :class:`ScheduledEvent`
    """

    def process(self):
        ret = ScheduledEvent(self.data)
        self.state.guild_events.new(ret.id, self.data)
        self.dispatch("scheduled_event", ret)


class OnScheduledEventUpdate(Event):
    """Gets called when a scheduled event is updated

    Returns
    -------
    before: :class:`ScheduledEvent`, can be :class:`None`
    after: :class:`SchduledEvent`
    """

    def process(self):
        after = ScheduledEvent(self.data)
        raw_before = self.state.guild_events.get(after.id)
        before = raw_before

        self.dispatch("scheduled_event_edit", before, after)
        self.state.guild_events.edit(after.id, self.data)


class OnScheduledEventDelete(Event):
    """Gets called when a scheduled event is deleted

    Returns
    -------
    Event: :class:`ScheduledEvent`
    """

    def process(self):
        event = ScheduledEvent(self.data)
        self.dispatch("scheduled_event_delete", event)
        self.state.guild_events.pop(event)


class OnScheduledEventJoin(Event):
    """Gets called when a scheduled event is joined by a member

    Returns
    -------
    Event: :class:`ScheduledEvent`
    User: :class:`User`
    Guild: :class:`Guild`
    """

    def process(self):
        raw_event_id = self.data["guild_scheduled_event_id"]
        raw_user_id = self.data["user_id"]
        raw_guild_id = self.data["guild_id"]
        raw_event = self.state.guild_events.get(raw_event_id)
        raw_user = self.state.members.get(raw_user_id)
        raw_guild = self.state.guilds.get(raw_guild_id)
        event = ScheduledEvent(raw_event)
        user = User(raw_user)
        guild = Guild(raw_guild)

        self.dispatch("scheduled_event_join", event, user, guild)


class OnScheduledEventLeave(Event):
    """Gets called when a user leaves a scheduled event

    Returns
    -------
    Event: :class:`ScheduledEvent`
    User: :class:`Member`
    Guild: :class:`Guild`
    """

    def process(self):
        raw_event_id = self.data["guild_scheduled_event_id"]
        raw_user_id = self.data["user_id"]
        raw_guild_id = self.data["guild_id"]
        raw_event = self.state.guild_events.get(raw_event_id)
        raw_user = self.state.members.get(raw_user_id)
        raw_guild = self.state.guilds.get(raw_guild_id)
        event = ScheduledEvent(raw_event)
        user = Member(raw_user)
        guild = Guild(raw_guild)

        self.dispatch("scheduled_event_join", event, user, guild)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""The V4 Voice Gateway Impl"""

import asyncio
import json
import logging
import struct
import time
from random import random

import aiohttp

from ..state import ConnectionState
from ..utils import create_snowflake

_log = logging.getLogger(__name__)


class VoiceGateway:
    def __init__(self, state: ConnectionState, guild_id: int, hook):
        self.state = state
        self.hook = hook
        self.server_id = guild_id
        self.session_id: int = None
        self.secret_key: str = None
        self.kept_alive: bool = False
        self.ws: aiohttp.ClientWebSocketResponse = None

    async def hook(self, *args):
        pass

    def update_session_id(self, id: int):
        self.session_id = id

    async def send_json(self, payload: dict):
        _log.debug("Sending payload data %s via the voice gateway", payload)
        await self.ws.send_str(json.dumps(payload))

    async def resume(self):
        payload = {
            "op": 7,
            "d": {
                "token": self.state.app.token,
                "server_id": str(self.server_id),
                "session_id": self.client.session_id,
            },
        }
        await self.send_json(payload)

    async def identify(self):
        payload = {
            "op": 0,
            "d": {
                "server_id": str(self.server_id),
                "user_id": self.state.app.user.id,
                "session_id": self.client.session_id,
                "token": self.state.app.token,
            },
        }
        await self.send_json(payload)

    async def connect(self, resume: bool = False):
        self.ws = await self.client._state.app.factory.ws_connect(
            self.gateway, compress=15
        )

        if not resume:
            await self.identify()
            self.client._state.loop.create_task(self.recv())
        else:
            await self.resume()
            self.client.start.loop.create_task(self.recv())

    # i know this is a loss of customization here but doing this makes,
    # the development experience so much easier and since i wouldn't guess people are customizing here
    # it's ok.
    @classmethod
    async def voice_client_entry(cls, client, *, hook=None):
        gateway = "wss://" + client.endpoint + "/?v=4"
        self = cls(client._state, client.guild_id, hook=hook)
        self.gateway = gateway
        self.client = client

        return self

    async def select_protocol(self, ip, port, mode):
        payload = {
            "op": 1,
            "d": {
                "protocol": "udp",
                "data": {
                    "address": ip,
                    "port": port,
                    "mode": mode,
                },
            },
        }
        await self.send_json(payload)

    async def client_connect(self):
        payload = {"op": 12, "d": {"audio_ssrc": self.client.ssrc}}
        await self.send_json(payload)

    async def speak(self, state=1):
        payload = {
            "op": 5,
            "d": {
                "speaking": int(state),
                "delay": 0,
            },
        }
        await self.send_json(payload)

    async def recv(self):
        async for msg in self.ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                data = json.loads(msg.data)
                op: int = data["op"]
                d: dict = data["d"]
                _log.debug("> %s", data)

                if op == 2:
                    await self.ready(d)

                elif op == 6:
                    self.kept_alive = True
                    _log.debug("> %s, kept the connection alive", d)

                elif op == 9:
                    _log.info("Resumed connection successfully")

                elif op == 4:
                    _log.info("Received session description")
                    self.client.mode = d["mode"]
                    self.secret_key = self.client.secret_key = d.get("secret_key")
                    await self.speak()
                    await self.speak(False)

                elif op == 8:
                    interval: int = d["heartbeat_interval"] / 1000
                    await self.hello(interval=interval)

    async def heartbeat(self, interval: float):
        while not self.ws.closed:
            if not self.kept_alive:
                await self.close(1008)
                await self.connect(resume=True)
            self.kept_alive = False
            await self.send_json({"op": 3, "d": create_snowflake()})
            await asyncio.sleep(interval)

    async def ready(self, data: dict):
        client = self.client
        client.ssrc = data["ssrc"]
        client.voice_port = data["port"]
        client.endpoint_ip = data["ip"]

        packet = bytearray(70)
        struct.pack_into(">H", packet, 0, 1)
        struct.pack_into(">H", packet, 2, 70)
        struct.pack_into(">I", packet, 4, client.ssrc)
        client.socket.sendto(packet, (client.endpoint_ip, client.voice_port))
        recv = self.state.loop.sock_recv(client.socket, 70)
        _log.debug("Received initial connection: %s", recv)

        ip_end = recv.index(0, 4)
        client.ip = recv[4:ip_end].decode("ascii")

        client.port = struct.unpack_from(">H", recv, len(recv) - 2)
        _log.debug("IP was detected: %s, port: %s", client.ip, client.port)

        modes = [mode for mode in data["modes"] if mode in self.client.supported_modes]
        _log.debug("Using following encryption mode(s): %s", modes)

        mode = modes[0]
        await self.select_protocol(client.ip, client.port, mode)
        _log.info("Using voice protocol: %s", mode)

    async def hello(self, interval: float):
        init = interval * random()
        await asyncio.sleep(init)
        self.kept_alive = (
            True  # exception to not kill the connection when saying hello.
        )
        await self.heartbeat(interval)

    async def close(self, code: int):
        await self.ws.close(code=code)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
import asyncio
import io
import logging
import threading
import time
import traceback
from typing import TYPE_CHECKING

_log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .client import VoiceClient


class AudioSource:
    def read(self) -> bytes:
        raise NotImplementedError

    def is_opus(self) -> bool:
        return False

    def cleanup(self) -> None:
        ...

    def __del__(self) -> None:
        return self.cleanup()


class PCMAudio(AudioSource):
    def __init__(self, stream: io.BufferedIOBase) -> None:
        self.stream = stream

    def read(self) -> bytes:
        ret = self.stream.read(20)
        if len(ret) != 20:
            return b""
        return ret


class AudioPlayer(threading.Thread):
    DELAY: float = 48000 / 1000

    def __init__(self, source: AudioSource, client, *, after=None):
        threading.Thread.__init__(self)
        self.daemon: bool = True
        self.source: AudioSource = source
        self.client: VoiceClient = client
        self.after = after

        # i would privatize these but to be fair
        # no one really uses this so ¯\_(ツ)_/¯
        self.end = threading.Event()
        self.resumed = threading.Event()
        self.resumed.set()
        self.error = None
        self.connected = self.client._connected
        self._lock = threading.Lock()

        if after is not None and not callable(after):
            raise TypeError()

    def _run(self):
        self.loops = 0
        self.Start = time.perf_counter()
        player = self.client.send_audio_packet
        self._speak(True)

        while not self.end.is_set():
            if not self.resumed.is_set():
                self.resumed.wait()
                continue

            if not self.connected.is_set():
                self.connected.wait()
                self.loops = 0
                self.start = time.perf_counter()

            self.loops += 1
            data = self.source.read()

            if not data:
                self.stop()
                break

            player(data, encode=not self.source.is_opus())
            next_time = self.start + self.DELAY * self.loops
            delay = max(0, self.DELAY + (next_time - time.perf_counter()))
            time.sleep(delay)

    def run(self):
        try:
            self._run()
        except Exception as exc:
            self.error = exc
            self.stop()
        finally:
            self.source.cleanup()
            self._call_after()

    def _call_after(self):
        error = self.error

        if self.after is not None:
            try:
                self.after(error)
            except Exception as exc:
                exc.__context__ = error
                traceback.print_exception(type(exc), exc, exc.__traceback__)
        elif error:
            _log.exception(f"Error in voice thread {self.name}")

    def stop(self):
        self.end.set()
        self.resumed.set()
        self._speak(False)

    def pause(self, *, update_speaking: bool = True):
        self.resumed.clear()
        if update_speaking:
            self._speak(False)

    def resume(self, *, update_speaking: bool = True):
        self.loops = 0
        self.start = time.perf_counter()
        self.resumed.set()
        if update_speaking:
            self._speak(True)

    def _set_source(self, source: AudioSource) -> None:
        with self._lock:
            self.pause(update_speaking=False)
            self.source = source
            self.resume(update_speaking=False)

    def _speak(self, speaking: bool) -> None:
        try:
            asyncio.run_coroutine_threadsafe(
                self.client.ws.speak(speaking), self.client._state.loop
            )
        except Exception as exc:
            _log.info("Speaking failed: %s", exc)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
from __future__ import annotations

from typing import Any
from typing import Dict as BaseDict
from typing import Type


def Dict() -> Type[BaseDict[str, Any]]:
    """Represents the widely used Dict format which discord.io uses.
        dict() -> new empty dictionary dict(mapping) -> new dictionary initialized from a mapping object's
        (key, value) pairs
    dict(iterable) -> new dictionary initialized as if via:
        d = {} for k, v in iterable:
            d[k] = v
    dict(**kwargs) -> new dictionary initialized with the name=value pairs
        in the keyword argument list. For example: dict(one=1, two=2)
    """
    return BaseDict[str, Any]



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE


from typing import Any, List

from ..assets import Emoji
from ..channels import TextChannel
from ..guild import Guild
from ..member import Member
from ..message import Message
from .core import Event

__all__: List[str] = ["OnMessage", "OnMessageEdit", "OnMessageDelete"]


class OnMessage(Event):
    """Returns a :class:`Message`"""

    def process(self) -> None:
        ret = Message(self.data, self.state.app)
        self.state.messages.new(self.data["id"], self.data)
        self.dispatch("MESSAGE", ret)


class OnMessageEdit(Event):
    """Returns the new :class:`Message` and if cached the old :class:`Message`"""

    def process(self):
        try:
            before = Message(self.state.messages.get(self.data["id"]), self.state.app)
        except KeyError:
            # if the message is not in the cache we cant really save it.
            before = None
        after = Message(self.data, self.state.app)
        self.dispatch("MESSAGE_EDIT", before, after)
        self.state.messages.edit(self.data["id"], self.data)


class OnMessageDelete(Event):
    """Gives the deleted :class:`Message`"""

    def process(self):
        message = Message(self.state.messages.pop(self.data["id"]), self.state.app)
        self.dispatch("MESSAGE_DELETE", message)


class OnMessageDeleteBulk(Event):
    """A event which gets called when multiple messages are deleted

    Returns
    -------
    Messages: list[:class:`Message`]
    Channel: :class:`TextChannel`
    Guild: :class:`Guild`
    """

    def process(self):
        msgs: list[dict[str, Any]] = [
            msg for msg in self.state.messages.get(self.data["ids"])
        ]
        messages = [Message(msg, self.state.app) for msg in msgs]
        channel = TextChannel(
            self.state.channels.get(self.data["channel_id"]), self.state
        )
        guild = Guild(
            self.state.guilds.get(self.data["guild_id"]), self.state.app.factory
        )

        self.dispatch("MESSAGE_BULK_DELETE", messages, channel, guild)

        for msg in self.data["ids"]:
            self.state.messages.pop(msg)


class OnMessageReactionAdd(Event):
    """Gives a :class:`Message` and a :class:`Emoji` that was added."""

    def process(self):
        emoji = Emoji(self.data["emoji"])
        member = Member(self.state.members.get(self.data["user_id"]))
        channel = TextChannel(self.state.channels.get(self.data["channel_id"]))
        message = Message(self.state.messages.get("message_id"))
        guild = Guild(self.state.guilds.get("guild_id"), self.state.app.factory)

        self.dispatch("MESSAGE_REACTION_ADD", emoji, message, channel, guild, member)


class OnMessageReactionRemove(Event):
    """Gives a :class:`Message` and a :class:`Emoji` that was removed."""

    def process(self):
        emoji = Emoji(self.data["emoji"])
        member = Member(self.state.members.get(self.data["user_id"]))
        channel = TextChannel(self.state.channels.get(self.data["channel_id"]))
        message = Message(self.state.messages.get("message_id"))
        guild = Guild(self.state.guilds.get("guild_id"), self.state.app.factory)

        self.dispatch("MESSAGE_REACTION_REMOVE", emoji, message, channel, guild, member)


class OnMessageReactionRemoveAll(Event):
    """Event called when all reactions are removed from a message

    Returns
    -------
    Channel: :class:`TextChannel`
    Message: :class:`Message`
    Guild: :class:`Guild`
    """

    def process(self):
        channel = TextChannel(self.state.channels.get(self.data["channel_id"]))
        message = Message(self.state.messages.get("message_id"))
        guild = Guild(self.state.guilds.get("guild_id"), self.state.app.factory)

        self.dispatch("MESSAGE_REACTION_REMOVE_ALL", message, channel, guild)


class OnMessageReactionRemoveEmoji(Event):
    """A event which happens when a emoji is removed from a guild and therefore the message

    Returns
    -------
    Emoji: :class:`Emoji`
    Channel: :class:`TextChannel`
    Message: :class:`Message`
    Guild: :class:`Guild`
    """

    def process(self):
        emoji = Emoji(self.data["emoji"])
        channel = TextChannel(self.state.channels.get(self.data["channel_id"]))
        message = Message(self.state.messages.get("message_id"))
        guild = Guild(self.state.guilds.get("guild_id"), self.state.app.factory)

        self.dispatch("MESSAGE_REACTION_REMOVE_EMOJI", emoji, message, channel, guild)



"""
discord.io
~~~~~~~~~~
Asynchronous Discord API Wrapper For Python

:copyright: 2021-present VincentRPS
:license: MIT
"""

__title__: str = "discord.io"
__author__: str = "VincentRPS"
__license__: str = "MIT"
__copyright__: str = "Copyright 2021-present VincentRPS"
__version__: str = "0.8.0"
__git_sha1__: str = "HEAD"

import logging
import typing

from .api import *
from .assets import *
from .channels import *
from .client import *
from .color import *
from .colour import *
from .components import *
from .embed import *
from .enums import *
from .events import *
from .file import *
from .flags import *
from .guild import *
from .intents import *
from .interactions import *
from .internal.exceptions import *
from .internal.warnings import *
from .member import *
from .message import *
from .snowflake import *
from .state import *
from .types import *
from .ui import *
from .user import *
from .utils import *
from .voice import *
from .webhooks import *


class VersionInfo(typing.NamedTuple):
    major: str
    minor: str
    micro: str
    releaselevel: typing.Literal["alpha", "beta", "candidate", "final"]
    serial: int


version_info: VersionInfo = VersionInfo(
    major=0, minor=8, micro=0, releaselevel="final", serial=0
)

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__: typing.List[str] = [
    "__title__",
    "__author__",
    "__license__",
    "__copyright__",
    "__version__",
    "VersionInfo",
    "version_info",
]



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

from typing import List, Optional, Union

from ..assets import Attachment
from ..channels import Category, TextChannel, Thread, VoiceChannel
from ..guild import Role
from ..member import Member
from ..user import User


def Choice(
    name: str,
    value: str,
):
    return {"name": name, "value": value}


def Option(
    name: str,
    description: str,
    type: Union[
        str,
        int,
        bool,
        User,
        TextChannel,
        VoiceChannel,
        Thread,
        Category,
        Role,
        Member,
        Attachment,
    ],
    required: Optional[bool] = False,
    choices: List[Choice] = None,
    channel_types: List[int] = None,
    min_value: int = None,
    max_value: int = None,
    autocomplete: bool = False,
):
    ret = {
        "name": name,
        "description": description,
    }

    # probably the most jank thing i wrote but it works, and does it good.
    if type == str:
        ret["type"] = 3
    elif type == int:
        ret["type"] = 4
    elif type == bool:
        ret["type"] = 5
    elif type == User:
        ret["type"] = 6
    elif type == TextChannel or VoiceChannel or Thread or Category:
        ret["type"] = 7
    elif type == Role:
        ret["type"] = 8
    elif type == Member:
        ret["type"] = 9
    elif type == Attachment:
        ret["type"] = 11

    if required is True:
        ret["required"] = True

    if choices is not None:
        ret["choices"] = choices

    if channel_types:
        ret["channel_types"] = channel_types

    if min_value:
        ret["min_value"] = min_value
    if max_value:
        ret["max_value"] = max_value

    if autocomplete:
        ret["autocomplete"] = True

    return ret



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

from typing import List

__all__: List[str] = ["Intents"]


class Intents:
    """Helps defining your Intents
    For a full list of intents and there usage please visit
    https://discord.dev/topics/gateway#gateway-intents

    .. versionadded:: 0.3.0

    Values
    ------
    default

    all

    GUILDS

    GUILD_MEMBERS

    GUILD_BANS

    GUILD_EMOJIS_AND_STICKERS

    GUILD_INTEGRATIONS

    GUILD_WEBHOOKS

    GUILD_INVITES

    GUILD_VOICE_STATES

    GUILD_PRESENCES

    GUILD_MESSAGES

    GUILD_MESSAGE_REACTIONS

    GUILD_MESSAGE_TYPING

    DIRECT_MESSAGES

    DIRECT_MESSAGE_REACTIONS

    DIRECT_MESSAGE_TYPING

    GUILD_SCHEDULED_EVENTS
    """

    def __init__(self, intents: int = 0):
        self.intents = intents
        if self.intents == "0":
            intents = self.default()

    @staticmethod
    def default() -> int:
        """Gives every non-privledged Intent

        Returns
        -------
        :class:`int`
        """
        return 32509

    @staticmethod
    def all() -> int:
        """Gives every Intent

        Returns
        -------
        :class:`int`
        """
        return 32767

    def GUILDS(self):
        return 1 << 0

    def GUILD_MEMBERS(self):
        return 1 << 1

    def GUILD_BANS(self):
        return 1 << 2

    def GUILD_EMOJIS_AND_STICKERS(self):
        return 1 << 3

    def GUILD_INTEGRATIONS(self):
        return 1 << 4

    def GUILD_WEBHOOKS(self):
        return 1 << 5

    def GUILD_INVITES(self):
        return 1 << 6

    def GUILD_VOICE_STATES(self):
        return 1 << 7

    def GUILD_PRESENCES(self):
        return 1 << 8

    def GUILD_MESSAGES(self):
        return 1 << 9

    def GUILD_MESSAGE_REACTIONS(self):
        return 1 << 10

    def GUILD_MESSAGE_TYPING(self):
        return 1 << 11

    def DIRECT_MESSAGES(self):
        return 1 << 12

    def DIRECT_MESSAGE_REACTIONS(self):
        return 1 << 13

    def DIRECT_MESSAGE_TYPING(self):
        return 1 << 14

    def GUILD_SCHEDULED_EVENTS(self):
        return 1 << 16



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
from typing import TYPE_CHECKING, Any, List, Optional

from discord.types import allowed_mentions

from ..api.rest import Route
from ..embed import Embed
from ..member import Member
from ..types import Dict, embed_parse
from ..webhooks import webhook_context

if TYPE_CHECKING:
    from ..state import ConnectionState

__all__: List[str] = ["Interaction"]


class Interaction:
    """Represents a Discord Interaction & Interaction Response

    Parameters
    ----------
    data
        The interaction data
    state
        The connection state

    Attributes
    ----------
    webhook
        A :class:`Webhook`
    token
        The interaction token
    type
        The interaction type
    guild_id
        The interaction guild id
    channel_id
        The interaction channel id
    data
        The interaction data
    id
        The interaction id
    message
        The interaction message

        .. note::

            This only appears on Component interactions.
    """

    def __init__(self, data: Dict, state):
        self.data = data
        self.state: ConnectionState = state
        self.collect_children(data)

    def collect_children(self, data):
        # collects the data
        self.token: str = data["token"]
        self.type: int = data["type"]
        self.guild_id: int = data["guild_id"]
        self.channel_id: int = data["channel_id"]
        self.data: Dict = data["data"]
        self.id: int = data["id"]

        try:
            self.options = data["data"]["options"]
        except KeyError:
            self.options = None

        try:
            self.modals = data["data"]["components"]
        except KeyError:
            self.modals = None

        try:
            # buttons will give this data
            self.message = data["message"]
        except KeyError:
            self.message = None

    def followup(
        self,
        content: Optional[str] = None,
        tts: bool = False,
        embed: Optional[Embed] = None,
        embeds: Optional[List[Embed]] = None,
    ):
        """Followup a defered interaction.

        Parameters
        ----------
        content
            The message content
        tts
            Text-To-Speach
        embed
            A :class:`Embed` object or :class:`dict`
        embeds
            A :class:`list` of :class:`Embed` or :class:`dict`
        allowed_mentions
            A allowed mentions object

        Returns
        -------
        :meth:`Webhook.execute`
        """
        adapter = webhook_context.get()
        return adapter.execute(
            id=self.state._bot_id,
            token=self.token,
            content=content,
            tts=tts,
            embed=embed,
            embeds=embeds,
            flags=self.invisable,
        )

    def respond(
        self,
        content: Optional[str] = None,
        modal: Optional[dict[str, Any]] = None,
        tts: bool = False,
        embed: Optional[Embed] = None,
        embeds: Optional[List[Embed]] = None,
        allowed_mentions: Optional[allowed_mentions.MentionObject] = None,
        type: Optional[int] = 4,
        invisable: Optional[bool] = False,
    ):
        """Send a initial response to a interaction

        Parameters
        ----------
        content
            The message content
        tts
            Text-To-Speach
        embed
            A :class:`Embed` object or :class:`dict`
        embeds
            A :class:`list` of :class:`Embed` or :class:`dict`
        allowed_mentions
            A allowed mentions object
        type
            The interaction type
        invisable
            If the interaction should only be seeable by the invoker.
        """
        ret = {"type": type, "data": {}}
        if content:
            ret["data"]["content"] = content
        ret["data"]["tts"] = tts
        if embed:
            emb = embed_parse.parse_embed(embed)
        if embeds:
            emb = embed_parse.parse_embeds(embeds)

        if not embeds and not embed:
            emb = []

        if allowed_mentions:
            ret["data"]["allowed_mentions"] = allowed_mentions
        else:
            ret["data"]["allowed_mentions"] = {"parse": []}

        if invisable:
            ret["data"]["flags"] = 1 << 6
            self.invisable = 1 << 6
        else:
            self.invisable = None

        ret["data"]["embeds"] = emb

        if modal:
            ret = {"type": 9, "data": modal}

        adapter = webhook_context.get()
        return adapter.rest.send(
            Route("POST", f"/interactions/{self.id}/{self.token}/callback"), json=ret
        )

    def defer(self, invisable: bool = False):
        """defers an interaction response

        Returns
        -------
        An empty :meth:`Interaction.respond`
        """

        return self.respond(type=5, invisable=invisable)

    @property
    def member(self):
        """Returns the member object of the invoker.

        Returns
        -------
        :class:`Member`
        """
        return Member(self.data["member"], self.state.app.factory)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""Implementation of Colors."""

from typing import List

__all__: List[str] = ["Color"]


class Color:
    """Represents the default discord colors

    Defines factory methods which return a certain color code to be used.

    .. versionadded:: 0.7.0
    """

    def __init__(self, value: int):
        if not isinstance(value, int):
            raise TypeError("Expected a integer.")

        self.value: int = value

    @classmethod
    def default(self) -> int:
        """A factory color method which returns `0`"""
        return int(0)

    @classmethod
    def teal(self) -> int:
        """A factory color method which returns `0x1ABC9C`"""
        return int(0x1ABC9C)

    @classmethod
    def dark_teal(self) -> int:
        """A factory color method which returns `0x11806A`"""
        return int(0x11806A)

    @classmethod
    def brand_green(self) -> int:
        """A factory color method which returns `0x57F287`"""
        return int(0x57F287)

    @classmethod
    def green(self) -> int:
        """A factory color method which returns `0x2ECC71`"""
        return int(0x2ECC71)

    @classmethod
    def dark_green(self) -> int:
        """A factory color method which returns `0x1F8B4C`"""
        return int(0x1F8B4C)

    @classmethod
    def blue(self) -> int:
        """A factory color method which returns `0x3498DB`"""
        return int(0x3498DB)

    @classmethod
    def dark_blue(self) -> int:
        """A factory color method which returns `0x206694`"""
        return int(0x206694)

    @classmethod
    def purple(self) -> int:
        """A factory color method which returns `0x9b59b6`"""
        return int(0x9B59B6)

    @classmethod
    def dark_purple(self) -> int:
        """A factory color method which returns `0x71368A`"""
        return int(0x71368A)

    @classmethod
    def magenta(self) -> int:
        """A factory color method which returns `0xE91E63`"""
        return int(0xE91E63)

    @classmethod
    def dark_magenta(self) -> int:
        """A factory color method which returns `0xAD1457`"""
        return int(0xAD1457)

    @classmethod
    def gold(self) -> int:
        """A factory color method which returns `0xF1C40F`"""
        return int(0xF1C40F)

    @classmethod
    def dark_gold(self) -> int:
        """A factory color method which returns `0xC27C0E`"""
        return int(0xC27C0E)

    @classmethod
    def orange(self) -> int:
        """A factory color method which returns `0xE67E22`"""
        return int(0xE67E22)

    @classmethod
    def dark_orange(self) -> int:
        """A factory color method which returns `0xA84300`"""
        return int(0xA84300)

    @classmethod
    def brand_red(self) -> int:
        """A factory color method which returns `0xED4245`"""
        return int(0xED4245)

    @classmethod
    def red(self) -> int:
        """A factory color method which returns `0xE74C3C`"""
        return int(0xE74C3C)

    @classmethod
    def dark_red(self) -> int:
        """A factory color method which returns `0x992D22`"""
        return int(0x992D22)

    @classmethod
    def dark_gray(self) -> int:
        """A factory color method which returns `0x607D8B`"""
        return int(0x607D8B)

    @classmethod
    def light_gray(self) -> int:
        """A factory color method which returns `0x979C9F`"""
        return int(0x979C9F)

    @classmethod
    def blurple(self) -> int:
        """A factory color method which returns `0x5865F2`"""
        return int(0x5865F2)

    @classmethod
    def dark_theme(self) -> int:
        """A factory color method which returns `0x2F3136`"""
        return int(0x2F3136)

    @classmethod
    def fushia(self) -> int:
        """A factory color method which returns `0xEB459E`"""
        return int(0xEB459E)

    @classmethod
    def yellow(self) -> int:
        """A factory color method which returns `0xFEE75C`"""
        return int(0xFEE75C)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
from discord.types import Dict

from ..guild import ScheduledEvent
from ..state import ConnectionState, member_cacher
from .channels import (
    OnChannelCreate,
    OnChannelDelete,
    OnChannelPinsUpdate,
    OnChannelUpdate,
    OnThreadCreate,
    OnThreadDelete,
    OnThreadListSync,
    OnThreadMembersUpdate,
    OnThreadMemberUpdate,
    OnThreadUpdate,
)
from .etc import (
    OnInviteCreate,
    OnInviteDelete,
    OnStageInstanceCreate,
    OnStageInstanceDelete,
    OnStageInstanceEdit,
    OnTyping,
    OnUserUpdate,
    OnWebhooksUpdate,
)
from .guilds import (
    OnGuildBan,
    OnGuildBanRemove,
    OnGuildEmojisUpdate,
    OnGuildIntegrationsUpdate,
    OnGuildJoin,
    OnGuildLeave,
    OnGuildStickersUpdate,
    OnGuildUpdate,
    OnMemberJoin,
    OnMemberLeave,
    OnMemberUpdate,
    OnRoleCreate,
    OnRoleDelete,
    OnRoleUpdate,
    OnScheduledEventCreate,
    OnScheduledEventDelete,
    OnScheduledEventJoin,
    OnScheduledEventLeave,
    OnScheduledEventUpdate,
)
from .interactions import OnInteraction
from .messages import (
    OnMessage,
    OnMessageDelete,
    OnMessageDeleteBulk,
    OnMessageEdit,
    OnMessageReactionAdd,
    OnMessageReactionRemove,
    OnMessageReactionRemoveAll,
    OnMessageReactionRemoveEmoji,
)


# https://discord.dev/topics/gateway#commands-and-events-gateway-events
class Cataloger:
    def __init__(self, data: Dict, dis, state: ConnectionState):

        # guilds
        if data["t"] == "GUILD_CREATE":
            state.guilds.new(data["d"]["id"], data["d"])
            # roles, channels
            for channel in data["d"]["channels"]:
                state.channels.new(channel["id"], channel)
            for role in data["d"]["roles"]:
                state.roles.new(role["id"], role)
            for event in data["d"]["guild_scheduled_events"]:
                even = ScheduledEvent(event)
                state.guild_events.new(even.id, even)
            dis.dispatch("RAW_GUILD_CREATE", data["d"])
            OnGuildJoin(data["d"], dis, state)

        elif data["t"] == "GUILD_DELETE":
            dis.dispatch("RAW_GUILD_DELETE", data["d"])
            OnGuildLeave(data["d"], dis, state)

        elif data["t"] == "GUILD_UPDATE":
            dis.dispatch("RAW_GUILD_UPDATE", data["d"])
            OnGuildUpdate(data["d"], dis, state)

        elif data["t"] == "GUILD_BAN_ADD":
            dis.dispatch("RAW_BAN_ADD", data["d"])
            OnGuildBan(data["d"], dis, state)

        elif data["t"] == "GUILD_BAN_REMOVE":
            dis.dispatch("RAW_BAN_REMOVE", data["d"])
            OnGuildBanRemove(data["d"], dis, state)

        elif data["t"] == "GUILD_EMOJIS_UPDATE":
            dis.dispatch("RAW_EMOJIS_UPDATE", data["d"])
            OnGuildEmojisUpdate(data["d"], dis, state)

        elif data["t"] == "GUILD_STICKERS_UPDATE":
            dis.dispatch("RAW_STICKERS_UPDATE", data["d"])
            OnGuildStickersUpdate(data["d"], dis, state)

        elif data["t"] == "GUILD_INTEGRATIONS_UPDATE":
            dis.dispatch("RAW_GUILD_INTEGRATIONS_UPDATE", data["d"])
            OnGuildIntegrationsUpdate(data["d"], dis, state)

        elif data["t"] == "GUILD_MEMBER_ADD":
            dis.dispatch("RAW_GUILD_MEMBER_ADD", data["d"])
            OnMemberJoin(data["d"], dis, state)

        elif data["t"] == "GUILD_MEMBER_REMOVE":
            dis.dispatch("RAW_GUILD_MEMBER_REMOVE", data["d"])
            OnMemberLeave(data["d"], dis, state)

        elif data["t"] == "GUILD_MEMBER_UPDATE":
            dis.dispatch("RAW_MEMBER_UPDATE", data["d"])
            OnMemberUpdate(data["d"], dis, state)

        elif data["t"] == "GUILD_MEMBERS_CHUNK":
            dis.dispatch("RAW_GUILD_MEMBERS_CHUNK", data["d"])
            member_cacher(state, data["d"]["members"])

        elif data["t"] == "ROLE_CREATE":
            dis.dispatch("RAW_ROLE_CREATE", data["d"])
            OnRoleCreate(data["d"], dis, state)

        elif data["t"] == "ROLE_UPDATE":
            dis.dispatch("RAW_ROLE_UPDATE", data["d"])
            OnRoleUpdate(data["d"], dis, state)

        elif data["t"] == "ROLE_DELETE":
            dis.dispatch("RAW_ROLE_DELETE", data["d"])
            OnRoleDelete(data["d"], dis, state)

        elif data["t"] == "GUILD_SCHEDULED_EVENT_CREATE":
            dis.dispatch("RAW_SCHEDULED_EVENT_CREATE", data["d"])
            OnScheduledEventCreate(data["d"], dis, state)

        elif data["t"] == "GUILD_SCHEDULED_EVENT_UPDATE":
            dis.dispatch("RAW_SCHEDULED_EVENT_UPDATE", data["d"])
            OnScheduledEventUpdate(data["d"], dis, state)

        elif data["t"] == "GUILD_SCHEDULED_EVENT_DELETE":
            dis.dispatch("RAW_SCHEDULED_EVENT_DELETE", data["d"])
            OnScheduledEventDelete(data["d"], dis, state)

        elif data["t"] == "GUILD_SCHEDULED_EVENT_USER_ADD":
            dis.dispatch("RAW_SCHEDULED_EVENT_JOIN", data["d"])
            OnScheduledEventJoin(data["d"], dis, state)

        elif data["t"] == "GUILD_SCHEDULED_EVENT_USER_REMOVE":
            dis.dispatch("RAW_SCHEDULED_EVENT_LEAVE", data["d"])
            OnScheduledEventLeave(data["d"], dis, state)

        # channels
        elif data["t"] == "CHANNEL_CREATE":
            dis.dispatch("RAW_CHANNEL_CREATE", data["d"])
            OnChannelCreate(data["d"], dis, state)

        elif data["t"] == "CHANNEL_UPDATE":
            dis.dispatch("RAW_CHANNEL_UPDATE", data["d"])
            OnChannelUpdate(data["d"], dis, state)

        elif data["t"] == "CHANNEL_DELETE":
            dis.dispatch("RAW_CHANNEL_DELETE", data["d"])
            OnChannelDelete(data["d"], dis, state)

        elif data["t"] == "CHANNEL_PINS_UPDATE":
            dis.dispatch("RAW_CHANNEL_PINS_UPDATE", data["t"])
            OnChannelPinsUpdate(data["d"], dis, state)

        elif data["t"] == "THREAD_CREATE":
            dis.dispatch("RAW_THREAD_CREATE", data["d"])
            OnThreadCreate(data["d"], dis, state)

        elif data["t"] == "THREAD_UPDATE":
            dis.dispatch("RAW_THREAD_UPDATE", data["d"])
            OnThreadUpdate(data["d"], dis, state)

        elif data["t"] == "THREAD_DELETE":
            dis.dispatch("RAW_THREAD_DELETE", data["d"])
            OnThreadDelete(data["d"], dis, state)

        elif data["t"] == "THREAD_LIST_SYNC":
            dis.dispatch("RAW_THREAD_LIST_SYNC", data["d"])
            OnThreadListSync(data["d"], dis, state)

        elif data["t"] == "THREAD_MEMBER_UPDATE":
            dis.dispatch("RAW_THREAD_MEMBER_UPDATE", data["d"])
            OnThreadMemberUpdate(data["d"], dis, state)

        elif data["t"] == "THREAD_MEMBERS_UPDATE":
            dis.dispatch("RAW_THREAD_MEMBERS_UPDATE", data["d"])
            OnThreadMembersUpdate(data["d"], dis, state)

        # messages
        elif data["t"] == "MESSAGE_CREATE":
            dis.dispatch("RAW_MESSAGE", data["d"])
            OnMessage(data["d"], dis, state)

        elif data["t"] == "MESSAGE_DELETE":
            dis.dispatch("RAW_MESSAGE_DELETE", data["d"])
            OnMessageDelete(data["d"], dis, state)

        elif data["t"] == "MESSAGE_UPDATE":
            dis.dispatch("RAW_MESSAGE_EDIT", data["d"])
            OnMessageEdit(data["d"], dis, state)

        elif data["t"] == "MESSAGE_DELETE_BULK":
            dis.dispatch("RAW_MESSAGE_DELETE_BULK", data["d"])
            OnMessageDeleteBulk(data["d"], dis, state)

        elif data["t"] == "MESSAGE_REACTION_ADD":
            dis.dispatch("RAW_MESSAGE_REACTION_ADD", data["d"])
            OnMessageReactionAdd(data["d"], dis, state)

        elif data["t"] == "MESSAGE_REACTION_REMOVE":
            dis.dispatch("RAW_MESSAGE_REACTION_REMOVE", data["d"])
            OnMessageReactionRemove(data["d"], dis, state)

        elif data["t"] == "MESSAGE_REACTION_REMOVE_ALL":
            dis.dispatch("RAW_MESSAGE_REACTION_REMOVE_ALL", data["d"])
            OnMessageReactionRemoveAll(data["d"], dis, state)

        elif data["t"] == "MESSAGE_REACTION_REMOVE_EMOJI":
            dis.dispatch("RAW_MESSAGE_REACTION_REMOVE_EMOJI", data["d"])
            OnMessageReactionRemoveEmoji(data["d"], dis, state)

        # interactions
        elif data["t"] == "INTERACTION_CREATE":
            dis.dispatch("RAW_INTERACTION", data["d"])
            OnInteraction(data["d"], dis, state)

        # etc
        elif data["t"] == "TYPING_START":
            dis.dispatch("RAW_TYPING_START", data["d"])
            OnTyping(data["d"], dis, state)

        elif data["t"] == "STAGE_INSTANCE_CREATE":
            dis.dispatch("RAW_STAGE_INSTANCE_CREATE", data["d"])
            OnStageInstanceCreate(data["d"], dis, state)

        elif data["t"] == "STAGE_INSTANCE_DELETE":
            dis.dispatch("RAW_STAGE_INSTANCE_DELETE", data["d"])
            OnStageInstanceDelete(data["d"], dis, state)

        elif data["t"] == "STAGE_INSTANCE_UPDATE":
            dis.dispatch("RAW_STAGE_INSTANCE_UPDATE", data["d"])
            OnStageInstanceEdit(data["d"], dis, state)

        elif data["t"] == "USER_UPDATE":
            dis.dispatch("RAW_USER_UPDATE", data["d"])
            OnUserUpdate(data["d"], dis, state)

        elif data["t"] == "WEBHOOKS_UPDATE":
            dis.dispatch("RAW_WEBHOOKS_UPDATE", data["d"])
            OnWebhooksUpdate(data["d"], dis, state)

        elif data["t"] == "INVITE_CREATE":
            dis.dispatch("RAW_INVITE_CREATE", data["d"])
            OnInviteCreate(data["d"], dis, state)

        elif data["t"] == "INVITE_DELETE":
            dis.dispatch("RAW_INVITE_DELETE", data["d"])
            OnInviteDelete(data["d"], dis, state)

        # there are some non-included events.
        else:
            dis.dispatch(f"RAW_{data['t']}", data["d"])



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""Represents a Discord Message

ref: https://discord.dev/resources/channel
"""

import asyncio
from typing import Any, List, Optional, Sequence

from discord.file import File
from discord.types import allowed_mentions

from .channels import TextChannel
from .embed import Embed
from .guild import Guild
from .user import User

__all__: List[str] = ["Message"]

# makes message data readable.
class Message:  # noqa: ignore
    """Represents a Discord Message

    .. versionadded:: 0.6.0

    Parameters
    ----------
    msg
        The message in dict format
    app
        The current :class:`Client`

    Attributes
    ----------
    from_dict
        A dict object of the message
    content
        The message content in string format
    channel
        The channel where the message happened
    """

    def __init__(self, msg: dict, app):
        self.from_dict = msg
        self.app = app
        try:
            self.content: str = self.from_dict["content"]
        except KeyError:
            # can error out for embed only/link only messages.
            self.content: str = ""

    def __repr__(self):
        return f"<Message id={self.id!r}, Channel id={self.channel.id!r}>"

    @property
    def channel(self):
        """Returns the channel which this message took place in.

        Returns
        -------
        :class:`TextChannel`
        """
        raw = self.app.state.channels.get(self.from_dict["channel_id"])
        return TextChannel(raw, self.app.state)

    @property
    def id(self) -> int:
        """Returns the message id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    @property
    def guild(self):
        """Returns the :class:`Guild` of the message

        Returns
        -------
        :class:`Guild`
        """
        return Guild(id=self.from_dict["guild_id"], rest_factory=self.app.factory)

    @property
    def author(self) -> User:
        """Returns the :class:`User` of the message

        Returns
        -------
        :class:`User`
        """
        return User(self.from_dict["author"])

    async def send(
        self,
        content: Optional[str] = None,
        files: Optional[Sequence[File]] = None,
        embed: Optional[Embed] = None,
        embeds: Optional[List[Embed]] = None,
        tts: Optional[bool] = False,
        allowed_mentions: Optional[allowed_mentions.MentionObject] = None,
        components: List[dict[str, Any]] = None,
        component=None,
    ):
        """Sends a message to the channel currently active in

        Parameters
        ----------
        content
            The message content
        files
            The message files
        embed
            A :class:`Embed`
        embeds
            A :class:`list` of :class:`Embed`
        tts
            A :class:`bool` of if tts should be on in this message
        allowed_mentions
            A allowed mentions object
        components
            A :class:`list` of components
        """
        emb = None
        com = None
        if embed and not embeds:
            if isinstance(embed, Embed):
                emb = [embed.obj]
            else:
                emb = [embed]
        elif embed and embed:
            raise TypeError("Used both `embed` and `embed` only 1 is allowed.")
        elif embeds and not embed:
            if isinstance(embeds, Embed):
                emb = [embed.obj for embed in embeds]
            else:
                emb = embeds
        if component:
            com = [component]
        if components:
            com = components
        await self.app.factory.create_message(
            channel=self.channel.id,
            content=content,
            files=files,
            embeds=emb,
            tts=tts,
            allowed_mentions=allowed_mentions,
            components=com,
        )

    async def reply(
        self,
        content: Optional[str] = None,
        files: Optional[Sequence[File]] = None,
        embed: Optional[Embed] = None,
        embeds: Optional[List[Embed]] = None,
        tts: Optional[bool] = False,
        allowed_mentions: Optional[allowed_mentions.MentionObject] = None,
        components: List[dict[str, Any]] = None,
        component: dict[str, Any] = None,
    ):
        """Replys to the current message

        Parameters
        ----------
        content
            The message content
        files
            The message files
        embed
            A :class:`Embed`
        embeds
            A :class:`list` of :class:`Embed`
        tts
            A :class:`bool` of if tts should be on in this message
        allowed_mentions
            A allowed mentions object
        components
            A :class:`list` of components
        """
        emb = None
        com = None
        if embed and not embeds:
            if isinstance(embed, Embed):
                emb = [embed.obj]
            else:
                emb = [embed]
        elif embed and embed:
            raise TypeError("Used both `embed` and `embed` only 1 is allowed.")
        elif embeds and not embed:
            if isinstance(embeds, Embed):
                emb = [embed.obj for embed in embeds]
            else:
                emb = embeds
        if component:
            com = [component]
        if components:
            com = components
        await self.app.factory.create_message(
            channel=self.channel.id,
            content=content,
            files=files,
            embeds=emb,
            tts=tts,
            allowed_mentions=allowed_mentions,
            message_reference={
                "message_id": self.id,
            },
            components=com,
        )



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""Represents a Discord Guild.

ref: https://discord.dev/resources/guild
"""
from typing import List, Optional, Union

from .assets import Emoji
from .enums import FormatType, ScheduledEventStatusType, ScheduledEventType
from .member import Member
from .user import User

__all__: List[str] = ["Guild"]


class Guild:
    """Represents a Discord Guild.

    .. versionadded:: 0.6.0

    Parameters
    ----------
    guild
        The raw guild object
    rest_factory
        The current :class:`RESTFactory`
    """

    # cache helpers for guilds.
    def __init__(self, guild: dict, rest_factory):
        self.from_dict = guild
        self._factory = rest_factory

    def joined_at(self) -> str:
        return self.from_dict["joined_at"]

    def vanity(self) -> str:
        """The vanity url, if None returns None

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["vanity_url_code"]

    def splash(self) -> str:
        """The splash screen, if None returns None

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["splash"]

    def discovery_splash(self) -> str:
        """The discovery splash, if None returns None

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["discovery_splash"]

    def sub_count(self) -> int:
        """The subscription count, returns a int

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["premium_subscription_count"]

    def emojis(self) -> List[Emoji]:
        """A list of :class:`Emoji`

        Returns
        -------
        list[:class:`Emoji`]
        """
        return [Emoji(emoji) for emoji in self.from_dict["emojis"]]

    @property
    def id(self) -> int:
        """The guild id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    async def get_member(self, id: int):
        """Gets a member and returns a :class:`Member` object

        Parameters
        ----------
        id
            The members id

        Returns
        -------
        :class:`Member`
        """
        unparsed = await self._factory.get_guild_member(self.id(), id)
        return Member(unparsed, self.id, self._factory)

    async def change_voice_state(
        self,
        *,
        channel: Optional[int] = None,
        self_mute: Optional[bool] = False,
        self_deaf: Optional[bool] = False,
    ):
        """Changes the bot's voice state in a guild

        .. versionadded:: 0.8.0

        Parameters
        ----------
        channel: :class:`int`
            The channel to connect to, None if to disconnect.
        self_mute
            Should the bot be muted?
        self_deaf
            Should the bot be deaf?
        """
        await self._factory.state.app.gateway.voice_state(
            guild=self.id, channel=channel, mute=self_mute, deaf=self_deaf
        )


def parse_role_icon(format: FormatType, role_id: int, role_icon: str) -> str:
    return f"https://cdn.discordapp.com/role-icons/{role_id}/{role_icon}.{format}"


def _parse_tags(data: dict) -> str:
    return data.get("bot_id") or data.get("integration_id") or ""


class Role:
    """Represents a Discord Role

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw role data
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def id(self) -> int:
        """Gives the snowflake id of the role

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    @property
    def name(self) -> str:
        """Gives the name of the role

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["name"]

    @property
    def color(self) -> int:
        """Returns the roles color

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["color"]

    def hoist(self) -> bool:
        """If the role is hoisted or not

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["hoist"]

    def icon(self, format: Optional[FormatType] = FormatType.PNG) -> str:
        """Gives the role icon's link, if any.

        Returns
        -------
        :class:`str`
        """
        return parse_role_icon(
            format=format, role_id=self.id, role_icon=self.from_dict["icon"]
        )

    def unicode_emoji(self) -> str:
        """Gives the roles unicode emoji

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["unicode_emoji"]

    @property
    def position(self) -> int:
        """Gives the current role position

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["position"]

    def permissions(self) -> str:
        """Gives the role permissions

        Returns
        -------
        :class:`str`
        """
        # TODO: Replace with a Permission class?
        return self.from_dict["permissions"]

    def managed(self) -> bool:
        """If the role is managed or not

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["managed"]

    def mentionable(self) -> bool:
        """If the role is publicly mentionable

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["mentionable"]

    def tags(self) -> str:
        """The current role' tags

        Returns
        -------
        :class:`str`
        """
        return _parse_tags(self.from_dict)


def parse_event_banner(format: FormatType, event_id: int, event_hash: str) -> str:
    return f"https://cdn.discordapp.com/guild-events/{event_id}/{event_hash}.{format}"


class ScheduledEvent:
    """Represents a Discord Guild Scheduled Event

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw event data
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def id(self) -> int:
        """Gives the scheduled events snowflake id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    def guild_id(self) -> int:
        """Gives the scheduled event' current guild

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["guild_id"]

    def channel_id(self) -> Union[int, None]:
        """Gives the scheduled events current channel, if any

        Returns
        -------
        :class:`int`
        :class:`None`
        """
        return self.from_dict.get("channel_id")

    @property
    def creator(self) -> User:
        """Gives a :class:`User` of the creator of this scheduled event

        Returns
        -------
        :class:`User`
        """
        return User(self.from_dict["creator"])

    @property
    def name(self) -> str:
        """Gives the scheduled event' name

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["name"]

    @property
    def description(self) -> str:
        """Gives the description of the scheduled event

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["description"]

    def start_time(self) -> str:
        """Gives the start time of the scheduled event

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["scheduled_start_time"]

    def status(self) -> ScheduledEventStatusType:
        """Gives the current event' status

        Returns
        -------
        :class:`int`
        """
        if self.from_dict["status"] == ScheduledEventStatusType.ACTIVE:
            return ScheduledEventStatusType.ACTIVE
        elif self.from_dict["status"] == ScheduledEventStatusType.COMPLETED:
            return ScheduledEventStatusType.COMPLETED
        elif self.from_dict["status"] == ScheduledEventStatusType.SCHEDULED:
            return ScheduledEventStatusType.SCHEDULED
        elif self.from_dict["status"] == ScheduledEventStatusType.CANCELED:
            return ScheduledEventStatusType.CANCELED

    def end_time(self) -> Union[None, str]:
        """Gives the endtime of the scheduled event

        Returns
        -------
        :class:`str`
        :class:`None`
        """
        return self.from_dict.get("scheduled_end_time")

    def type(self) -> ScheduledEventType:
        """Gives the type of scheduled event

        Returns
        -------
        :class:`int`
        """
        if self.from_dict["entity_type"] == ScheduledEventType.STAGE_INSTANCE:
            return ScheduledEventType.STAGE_INSTANCE
        elif self.from_dict["entity_type"] == ScheduledEventType.VOICE:
            return ScheduledEventType.VOICE
        elif self.from_dict["entity_type"] == ScheduledEventType.EXTERNAL:
            return ScheduledEventType.EXTERNAL

    def entity_id(self) -> int:
        """Gives the current entity id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["entity_id"]

    @property
    def metadata(self) -> "ScheduledEventMetadata":
        """Gives the scheduled event metadata

        Returns
        -------
        :class:`ScheduledEventMetadata`
        """
        return ScheduledEventMetadata(self.from_dict["entity_metadata"])

    def joined(self) -> int:
        """Gives the amount of users which joined the event

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["user_count"]

    def image(self, format: FormatType = FormatType.PNG):
        """Gives the url of the scheduled event' banner

        Returns
        -------
        :class:`str`
        """
        return parse_event_banner(
            format=format, event_id=self.id, event_hash=self.from_dict["image"]
        )


class ScheduledEventMetadata:
    """Represents a Discord Scheduled event metadata

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw metadata' data
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def location(self) -> Union[str, None]:
        """Gives the location the event is happening in

        Returns
        -------
        :class:`str`
        :class:`None`
        """
        return self.from_dict["location"]


class WelcomeScreen:
    """Represents a Discord Guild WelcomeScreen

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def description(self) -> str:
        """Gives the WelcomeScreen' description

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["description"]

    def channels(self) -> list["WelcomeChannel"]:
        """Gives a list of :class:`WelcomeChannel`

        Returns
        -------
        list[:class:`WelcomeChannel`]
        """
        return [
            WelcomeChannel(channel) for channel in self.from_dict["welcome_channels"]
        ]


class WelcomeChannel:
    """Represents a Discord WelcomeScreen Channel

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw WelcomeChannel data
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def channel_id(self) -> int:
        """Gives the Channel' id


        Returns
        -------
        :class:`int`
        """
        return self.from_dict["channel_id"]

    @property
    def description(self) -> str:
        """Gives the description of the channel

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["description"]

    def emoji_id(self) -> Union[int, None]:
        """Gives the Emoji id, if any

        Returns
        -------
        :class:`int`
        :class:`None`
        """
        return self.from_dict.get("emoji_id")

    def emoji_name(self) -> Union[str, None]:
        """Gives the Emoji name, if any

        Returns
        -------
        :class:`str`
        :class:`None`
        """
        return self.from_dict.get("emoji_name")



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""
Dispatches raw Opcode events to Event classes.
"""

import asyncio
import logging
from typing import Any, Callable, Coroutine, List, Optional, TypeVar

from discord.state import ConnectionState

from ..ext import cogs

__all__: List[str] = ["Dispatcher"]
_log = logging.getLogger(__name__)
CoroT = TypeVar("CoroT", bound=Callable[..., Coroutine[Any, Any, Any]])
T = TypeVar("T")
Coro = Coroutine[Any, Any, T]
CoroFunc = Callable[..., Coro[Any]]


class Dispatcher:
    """Dispatches raw and non-raw events

    Parameters
    ----------
    state
        The :class:`ConnectionState`
    """

    def __init__(self, state: ConnectionState):
        self.state = state

    async def run(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        name: str,
        cog,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        try:
            if cog:
                await coro(cog, *args, **kwargs)
            else:
                await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception:
            raise

    def scheduler(
        self,
        coro: Callable[..., Coroutine[Any, Any, Any]],
        name: str,
        cog: Any,
        *args: Any,
        **kwargs: Any,
    ) -> asyncio.Task:
        wrap = self.run(coro, name, cog, *args, **kwargs)
        return self.state.loop.create_task(wrap, name=f"discord.io: {name}")

    def dispatch(self, name: str, *args, **kwargs) -> None:
        fake_name = str(name.lower())
        real_name = "on_" + str(fake_name)
        _log.debug("Dispatching event: %s", real_name)

        listeners = self.state.listeners.get(real_name)
        if listeners:
            removed = []
            for i, (future, condition) in enumerate(listeners):
                if future.cancelled():
                    removed.append(i)
                    continue

                try:
                    result = condition(*args)
                except Exception as exc:
                    future.set_exception(exc)
                    removed.append(i)
                else:
                    if result:
                        if len(args) == 0:
                            future.set_result(None)
                        elif len(args) == 1:
                            future.set_result(args[0])
                        else:
                            future.set_result(args)
                        removed.append(i)

            if len(removed) == len(listeners):
                self.state.listeners.pop(real_name)
            else:
                for res in reversed(listeners):
                    removed.append(res)

        try:
            coro = getattr(self, real_name)
        except AttributeError:
            ...
        else:
            self.scheduler(coro["main"], real_name, coro["cog"], *args, **kwargs)

    def listen(self, coro: Coro) -> Coro:
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("Function is not a coroutine.")

        setattr(self, coro.__name__, coro)
        _log.info(f"{coro.__name__} has been registered!")

    def add_listener(self, func: CoroFunc, name: Optional[str] = None, cog=None):
        name = func.__name__ if name is None else name

        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Function is not a coroutine.")

        setattr(self, name, {"main": func, "cog": cog})

        if name.startswith("on_raw"):
            _log.debug(f"{name} added as a listener!")
        else:
            _log.info(f"{name} added as a listener!")

    def remove_listener(self, func: CoroFunc, name: Optional[str] = None):
        name = func.__name__ if name is None else name

        if name in self.state.listeners:
            try:
                self.state.listeners[name].remove(func)
            except ValueError:
                pass



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""Represents a Discord Member

ref: https://discord.dev/resources/guild#guild-member-object
"""
from typing import Any, List, Optional, Union

from .api.rest_factory import RESTFactory
from .user import User

__all__: List[str] = ["Member"]


class Member:
    """Represents a Discord Guild Member

    .. versionadded:: 0.7.0

    Parameters
    ----------
    data
        The member data, a :class:`dict`
    factory
        The current instance of :class:`RESTFactory`

    Attributes
    ----------
    from_dict
        The raw :class:`dict` object of the Member.
    """

    def __init__(self, data: dict, guild: Union[int, Any], factory: RESTFactory):
        self.from_dict = data
        self.guild_id = guild
        self._factory = factory

    @property
    def user(self):
        """Returns the members :class:`User` object

        Returns
        -------
        :class:`User`
        """
        return User(self.from_dict["user"])

    def nick(self) -> str:
        """Returns the members nick name, if any

        Returns
        -------
        :class:`str`
        :class:`None`
        """
        return self.from_dict["nick"]

    def avatar(self) -> str:
        raise NotImplementedError

    def roles(self):
        # i need to somehow make this not be a circular import?
        raise NotImplementedError

    def joined_at(self) -> str:
        """Gives a timestamp of when the member joined the server

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["joined_at"]

    def premium_since(self) -> str:
        """Gives a timestamp of when the member started boosting

        Returns
        -------
        :class:`str`
        :class:`None`
        """
        return self.from_dict["premium_since"]

    def deaf(self) -> bool:
        """Returns a bool if the member is deaf in a voice channel or not.

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["deaf"]

    def mute(self) -> bool:
        """Returns if the member it muted from a channel

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["mute"]

    def pending(self) -> bool:
        """Returns if the user is pending verification or not

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["pending"]

    def permissions(self) -> dict[str, Any]:
        """Returns a dict of the users permissions

        Returns
        -------
        :class:`dict`
        """
        return self.from_dict["permissions"]

    def communication_disabled_until(self) -> str:
        """Gives the time communication is disabled

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["communication_disabled_until"]

    def edit(
        self,
        nick: Optional[str] = None,
        roles: Optional[List[int]] = None,
        mute: Optional[bool] = False,
        deaf: Optional[bool] = False,
        channel_id: Optional[int] = None,
        timeout: Optional[str] = None,
        reason: Optional[str] = None,
    ) -> None:
        """Edits the member

        Parameters
        ----------
        nick: Optional[:class:`str`]
            Change the members nickname
        roles: Optional[:class:`list`[:class:`int`]]
            Chaneg the members roles
        mute: Optional[:class:`bool`]
            If the member should be muted
        deaf: Optional[:class:`bool`]
            If the member should be deafend
        channel_id: Optional[:class:`int`]
            The channel id to move the member to
        timeout: Optional[:class:`str`]
            Set a timeout for the member,
            has to be a ISO8601 timestanp
        reason: Optional[:class:`str`]
            A reason why you are editing this member

        Returns
        -------
        :class:`None`
        :class:`Forbidden`
        """
        return self._factory.modify_guild_member(
            guild_id=self.guild_id,
            member=self.user.id,
            nick=nick,
            roles=roles,
            mute=mute,
            deaf=deaf,
            channel_id=channel_id,
            timeout=timeout,
            reason=reason,
        )


class Presence:
    """Represents a Discord Presence

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw presence data
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def user(self) -> User:
        """Gives the :class:`User` which has this presence

        Returns
        -------
        :class:`User`
        """
        return User(self.from_dict["user"])

    @property
    def guild_id(self) -> int:
        """Gives the id of the guild

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["guild_id"]

    @property
    def status(self) -> str:
        """The users current status, i.e 'online', 'offline', etc

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["status"]

    @property
    def activites(self) -> List["Activity"]:
        """Gives the users activites

        Returns
        -------
        List[:class:`Activity`]
        """
        return [Activity(raw) for raw in self.from_dict["activites"]]

    @property
    def client_status(self) -> str:
        """The users current client status, i.e desktop, mobile, etc

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["client_status"]


# still needs finishing i just don't feel like doing it now.
class Activity:
    """Represents a Discord Activity

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw activity data
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def name(self) -> str:
        return self.from_dict["name"]

    @property
    def type(self) -> int:
        return self.from_dict["type"]

    @property
    def url(self) -> Union[str, None]:
        return self.from_dict["url"]



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""Implementation of the Discord Gateway."""

from __future__ import annotations

import asyncio
import json
import logging
import platform
import zlib
from random import random
from time import perf_counter, time
from typing import Any, Coroutine, List

import aiohttp

from discord import utils
from discord.events import catalog
from discord.internal.dispatcher import Dispatcher
from discord.snowflake import Snowflakeish
from discord.types.dict import Dict

from ..state import ConnectionState
from .rest_factory import RESTFactory

ZLIB_SUFFIX = b"\x00\x00\xff\xff"
_log = logging.getLogger(__name__)
url_extension = "?v=9&encoding=json&compress=zlib-stream"


class Shard:
    """Represents a Discord Shard.

    Parameters
    ----------
    state
        The :class:`ConnectionState` cache to use.
    dispatcher
        The :class:`Dispatcher` to use.
    shard_id: :class:`int`
        The shard number.
    url: :class:`str`
        The URL to use.
    mobile :class:`bool`
        If to have a mobile presence or not.

    Attributes
    ----------
    buffer
        An array of bytes which buffers the connection.
    remaining
        The amount remaining
    per
        The times per
    window
        The window
    max
        The max
    inflator
        The inflator for this shard.
    """

    def __init__(
        self,
        state: ConnectionState,
        dispatcher: Dispatcher,
        shard_id: int,
        url: str,
        mobile: bool = False,
    ):
        self.remaining = 110
        self.per = 60.0
        self.window = 0.0
        self.max = 110
        self.state = state
        self.url = url
        self.mobile = mobile
        self.dis = dispatcher
        self.inflator = zlib.decompressobj()
        self.shard_id = shard_id
        self.buffer = bytearray()
        self.last_recv = perf_counter()
        self.last_send = perf_counter()
        self.last_ack = perf_counter()
        self._session_id: int = None
        self._ratelimit_lock: asyncio.Lock = asyncio.Lock()
        self.ws: aiohttp.ClientWebSocketResponse = None
        self.latency: float = float("nan")

    async def connect(self, token: str) -> None:
        """Connects to the url specified, with the token

        Parameters
        ----------
        token
            Your bot token.

        Attributes
        ----------
        _session
            The aiohttp ClientSession
        ws
            The WebSocket connection.
        """
        self._session = aiohttp.ClientSession()
        self.ws = await self._session.ws_connect(self.url)

        self.token = token
        if self._session_id is None:
            await self.identify()
            self.state.loop.create_task(self.recv())
            self.state.loop.create_task(self.check_connection())
        else:
            await self.resume()
            _log.debug("Reconnected to the Gateway")

    @property
    def is_ratelimited(self) -> bool:
        """Returns a True if this shard is ratelimited
        and False if it isn't
        """
        now = time()
        if now > self.window + self.per:
            return False
        return self.remaining == 0

    def delay(self) -> float:
        """A float showing how long we should delay until retrying."""
        now = time()

        if now > self.window + self.per:
            self.remaining = self.max

        if self.remaining == self.max:
            self.window = now

        if self.remaining == 0:
            return self.per - (now - self.window)

        self.remaining -= 1
        if self.remaining == 0:
            self.window = now

        return 0.0

    async def block(self) -> None:
        """A function to block the connection tempoarily"""
        async with self._ratelimit_lock:
            delay = self.delay()
            if delay:
                _log.warning(
                    "Shard %s was ratelimited, waiting %.2f seconds.",
                    self.shard_id,
                    delay,
                )
                await asyncio.sleep(delay)

    async def check_connection(self):
        await asyncio.sleep(50)
        if self.last_recv + 60.0 < perf_counter():
            _log.warning(
                f"Shard {self.shard_id} has stopped receiving from the gateway, reconnecting"
            )
            await self.ws.close(code=4000)
            await self.closed(4000)
        elif self.latency > 10:
            _log.warning(f"Shard {self.shard_id} is behind by {self.latency}")
        self.state.loop.create_task(self.check_connection())

    async def send(self, data: Dict) -> None:
        """Send a request to the Gateway via the shard

        Parameters
        ----------
        data :class:`Dict`
            The data to send
        """
        self.last_send = perf_counter()
        _log.debug("< %s", data)
        raw_payload = json.dumps(data)

        if isinstance(raw_payload, str):
            payload = raw_payload.encode("utf-8")

        await self.ws.send_bytes(payload)

    async def recv(self) -> None:
        async for msg in self.ws:
            if msg.type == aiohttp.WSMsgType.BINARY:
                self.latency = self.last_ack - self.last_send
                self.buffer.extend(msg.data)
                try:
                    raw = self.inflator.decompress(self.buffer).decode("utf-8")
                except:
                    # probably corrupted data
                    self.buffer = bytearray()
                    return
                if len(msg.data) < 4 or msg.data[-4:] != ZLIB_SUFFIX:
                    raise RuntimeError
                self.buffer = bytearray()  # clean buffer
                data = json.loads(raw)
                _log.debug("> %s", data)

                self._seq = data["s"]
                self.dis.dispatch("RAW_SOCKET_RECEIVE")

                self.last_recv = perf_counter()

                if data["op"] == 0:
                    if (
                        data["t"] == "READY"
                    ):  # only fire up getting the session_id on a ready event.
                        await self._ready(data)
                        self.dis.dispatch("READY")
                    else:
                        catalog.Cataloger(data, self.dis, self.state)
                elif data["op"] == 9:
                    await self.ws.close(code=4000)
                    await self.closed(4000)
                elif data["op"] == 10:
                    await self.hello(data)
                elif data["op"] == 11:
                    self.last_ack = perf_counter()
                    _log.debug("> %s", data)
                else:
                    self.dis.dispatch(data["op"], data)

            else:
                raise RuntimeError

        code = self.ws.close_code
        if code is None:
            return
        else:
            await self.closed(code)

    async def closed(self, code: int) -> None:
        _log.error(f"Shard {self.shard_id} closed with code {code}")
        if code == 4000:
            pass

        elif code == 4001:
            pass

        elif code == 4002:
            # just ignore it.
            pass

        elif code == 4003:
            # something weird happened, just ignore it.
            pass

        elif code == 4004:
            raise

        elif code == 4005:
            # pass!
            pass

        elif code == 4007:
            # pass and make a new session
            pass

        elif code == 4008:
            # retry!
            pass

        elif code == 4009:
            # try to resume.
            await self.resume()
            return

        elif code == 4010:
            raise  # this doesn't really happen, unless you are dumb

        elif code == 4011:
            # just tell the owner to shard!
            raise

        elif code == 4012:
            # most likely a old lib version.
            raise

        elif code == 4013:
            raise

        elif code == 4014:
            raise

        await self.connect(token=self.token)

    async def heartbeat(self, interval: float) -> None:
        if self.is_ratelimited:
            await self.block()
        if not self.ws.closed:
            await self.send({"op": 1, "d": self._seq})
            await asyncio.sleep(interval)
            self.state.loop.create_task(self.heartbeat(interval))

    async def close(self, code: int = 4000) -> None:
        if self.ws:
            await self.ws.close(code=code)
        self.buffer.clear()

    async def hello(self, data: Dict) -> None:
        interval = data["d"]["heartbeat_interval"] / 1000
        init = interval * random()
        await asyncio.sleep(init)
        self.state.loop.create_task(self.heartbeat(interval))

    async def _ready(self, data: Dict) -> None:
        self._session_id = data["d"]["session_id"]
        self.state._ready.set()

    async def identify(self) -> None:
        await self.send(
            {
                "op": 2,
                "d": {
                    "token": self.token,
                    "intents": self.state._bot_intents,
                    "properties": {
                        "$os": platform.system(),
                        "$browser": "discord.io"
                        if self.mobile is False
                        else "Discord iOS",
                        "$device": "discord.io",
                    },
                    "shard": (self.shard_id, self.state.shard_count),
                    "v": 9,
                    "compress": True,
                },
            }
        )

    async def resume(self) -> None:
        await self.send(
            {
                "op": 6,
                "d": {
                    "token": self.token,
                    "session_id": self._session_id,
                    "seq": self._seq,
                },
            }
        )


class Gateway:
    """Represents a Gateway connection with Discord.

    Parameters
    ----------
    state
        The :class:`ConnectionState` to use.
    dispatcher
        The :class:`Dispatcher` to use.
    factory
        The :class:`RESTFactory` to use.
    mobile
        If your bot should have a mobile presence or not.
    """

    def __init__(
        self,
        state: ConnectionState,
        dispatcher: Dispatcher,
        factory: RESTFactory,
        mobile: bool = False,
    ):
        self._s = state
        self.mobile = mobile
        self.count = self._s.shard_count
        self._d = dispatcher
        self._f = factory
        self.shards: List[Shard] = []

    @utils.copy_doc(Shard.connect)
    async def connect(self, token: str) -> None:
        r = await self._f.get_gateway_bot()
        if self.count is None:
            shds = r["shards"]
            self._s.shard_count = shds
        else:
            shds = self.count

        for shard in range(shds):
            self.s = Shard(
                self._s,
                self._d,
                shard,
                mobile=self.mobile,
                url=r["url"] + url_extension,
            )
            self._s.loop.create_task(self.s.connect(token))
            self.shards.append(self.s)
            _log.info("Shard %s has connected to Discord", shard)

    @utils.copy_doc(Shard.send)
    def send(self, payload: Dict) -> Coroutine[Any, Any, None]:
        return self.s.send(payload)

    async def _chunk_members(self):
        await asyncio.sleep(20)
        for guild in self._s.guilds._cache.values():
            shard_id = (int(guild["id"]) >> 22) % self._s.shard_count
            await self.shards[shard_id].send(
                {"op": 8, "d": {"guild_id": guild["id"], "query": "", "limit": 0}}
            )

    async def voice_state(
        self, guild: int, channel: Snowflakeish, mute: bool, deaf: bool
    ) -> Coroutine[Any, Any, None]:
        json = {
            "op": 4,
            "d": {
                "guild_id": guild,
                "channel_id": channel,
                "self_mute": mute,
                "self_deaf": deaf,
            },
        }
        shard_id = (int(guild) >> 22) % self._s.shard_count
        await self.shards[shard_id].send(json)

    @property
    def latency(self) -> float:
        lat: float = float(0)
        for shard in self.shards:
            lat += shard.latency

        return lat



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE
"""
The ConnectionState Caches most things during connection.
"""
import asyncio
from collections import OrderedDict
from typing import Any, Callable, Coroutine, List, Tuple, TypeVar, Union

from discord.types.dict import Dict

__all__: List = ["Hold", "ConnectionState"]

T = TypeVar("T")
Coro = Coroutine[Any, Any, T]
CoroFunc = Callable[..., Coro[Any]]


# so there are 2 pretty big problems with this right now,
# 1 it's not async so the db would have to be blocking
# 2 it's repetitive to switch out since you need to switch all cache objects.
class Hold:
    """A hold of cache, easily swapable with a db."""

    def __init__(self):
        self._cache = OrderedDict()

    def view(self) -> List[Dict]:
        self._cache.values()

    def list(self):
        return self._cache.items()

    def new(self, name: str, data: Union[str, int, Dict, Any]):
        self._cache[name] = data

    def edit(self, name: str, data: Union[str, int, Dict, Any]):
        self._cache.update({name: data})

    def get(self, name: str):
        return self._cache.get(name)

    def pop(self, name: str):
        return self._cache.pop(name)

    def reset(self) -> None:
        del self._cache


class ConnectionState:
    """The Connection State

    .. versionadded:: 0.4.0

    .. note::

        The connection state is responsible for caching
        everything, meaning most classes will depend on it.

    .. note::

        .. data:: ConnectionState._speaking

        is planned to be deprecated soon.

    Attributes
    ----------
    _bot_intents: :class:`int`
        The cached bot intents, used for Gateway

    _session_id: :class:`int`
        The Gateway, session id

    _voice_session_id: :class:`int`
        The Voice Gateway Session ID

    _seq: :class:`int`
        The Gateway seq number, can be None.

    app: :class:`Client`
        The bot app

        .. versionadded:: 0.5.0

    _voice_seq: :class:`int`
        The Voice Gateway seq

        .. versionadded:: 0.5.0

    _voice_user_data: :class:`dict`
        The Voice User Data given by the Voice Gateway.

        .. versionadded:: 0.5.0

    _speaking: :class:`bool`
        If the bot is currently speaking, defaults False

        .. versionadded:: 0.5.0

    _said_hello: :class:`bool`
        If the Gateway got a hello or not.

    loop :class:`asyncio.AbstractEventLoop`
        The current loop

    _bot_presences: :class:`list`
        A list of the bots presences

    _bot_status: :class:`str`
        The bot status, e.g. online

    _bot_presence_type: :class:`int`
        The bot presence type, defaults to 0

    listeners: :class:`dict`
        The bot listeners

    shard_count: :class:`int`
        the number of shards.

        .. versionadded:: 0.6.0
    """

    def __init__(self, **options):
        self.guilds = options.get("guild_cache_hold") or Hold()
        self.members = options.get("members_cache_hold") or Hold()
        self.roles = options.get("roles_cache_hold") or Hold()
        self.guild_events = options.get("guild_events_cache_hold") or Hold()
        self.channels = options.get("channel_cache_hold") or Hold()
        self.messages = options.get("messages_cache_hold") or Hold()
        self.stage_instances = options.get("stage_instances_cache_hold") or Hold()
        self._ready: asyncio.Event = asyncio.Event()

        self._bot_intents: int = options.get("intents", 0)
        """The cached bot intents, used for Gateway"""

        self._bot_id: int = None

        self.bot_info = {}

        self._voice_session_id: int = None

        self._seq: int = None
        """The seq number"""

        self.app = options.get("bot", None)
        """The bot app"""

        self._voice_seq: int = None
        """The Voice Gateway seq"""

        self._voice_user_data: Dict = {}
        """The Voice User Data given by the Voice Gateway."""

        self._speaking: bool = False
        """If the bot is currently speaking"""

        self._said_hello: bool = False
        """If the Gateway got a hello or not."""

        self.loop: asyncio.AbstractEventLoop = options.get("loop", None)
        """The current loop"""

        self._bot_presences: list[str, Any] = []
        """The precenses"""

        self._bot_status: str = "online"
        """The status"""

        self._bot_presence_type: int = 0
        """Precense type"""

        self.listeners: dict[str, List[Tuple[asyncio.Future, Callable[..., bool]]]] = {}
        """The listeners"""

        self.shard_count: int = options.get("shard_count", None)
        """The shard count"""

        self.components: dict[str, Any] = {}

        self.prefixed_commands: dict[str, List[CoroFunc]] = {}

        self.application_commands: dict[str, Any] = {}

        self.prefix = options.get("prefix")


def member_cacher(state: ConnectionState, data: Any):
    for member in data:
        state.members.new(member["id"], member)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

from typing import Union

from .enums import FormatType
from .guild import Guild
from .state import ConnectionState
from .user import User


def channel_parse(type, data: dict, state: ConnectionState):
    if type == 0:
        return TextChannel(data, state)

    elif type == 1:
        return DMChannel(data, state)

    elif type == 2:
        return VoiceChannel(data, state)

    elif type == 3:
        return GroupDMChannel(data, state)

    elif type == 4:
        return Category(data, state)

    elif type == 5:
        return TextChannel(data, state)

    elif type in (10, 11, 12):
        return Thread(data, state)

    elif type == 13:
        return VoiceChannel(data, state)

    else:
        raise NotImplementedError("Channel is not a currently provided type")


class Category:
    """Represents a Discord Category

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw category data
    state: :class:`state`
        The connection state
    """

    def __init__(self, data: dict, state: ConnectionState):
        self.from_dict = data
        self.state = state

    def permission_overwrites(self) -> list[str, Union[int, str]]:
        """Gives a list of permission overwrites

        Returns
        -------
        list[:class:`str`]
        list[:class:`int`]
        """
        return self.from_dict["permission_overwrites"]

    @property
    def position(self) -> int:
        """Gives the category position

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["position"]

    @property
    def id(self) -> int:
        """Gives the snowflake id of the channel

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    @property
    def name(self) -> str:
        """Gives the name of the category

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["name"]

    def guild_id(self) -> int:
        """Gives the guild' snowflake id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["guild_id"]


class TextChannel:
    """Represents a Discord Text Channel

    Parameters
    ----------
    data: :class:`dict`
        The raw json data
    state: :class:`ConnectionState`
        The connection state
    """

    def __init__(self, data: dict, state: ConnectionState):
        self.from_dict = data
        self.state = state

    @property
    def id(self) -> int:
        """Gives the channel' id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    @property
    def guild_id(self) -> int:
        """Gives the guild id of the channel

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["guild_id"]

    @property
    def name(self) -> str:
        """Gives the name of the channel

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["name"]

    @property
    def position(self) -> int:
        """Gives the position of the channel

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["position"]

    def permission_overwrites(self) -> list[str, Union[int, str]]:
        """Gives the permission overwrites of the channel"""
        return self.from_dict["permission_overwrites"]

    @property
    def nsfw(self) -> bool:
        """If the channel is nsfw

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["nsfw"]

    def topic(self) -> str:
        """Gives the channel' topic

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["topic"]

    def last_message_id(self) -> int:
        """Gives the snowflake id of the last message

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["last_message_id"]

    def category_id(self) -> int:
        """Gives the id of the category this channel is in

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["parent_id"]


class VoiceChannel:
    """Represents a Discord Voice Channel

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw channel data
    state: :class:`ConnectionState`
        The connection state
    """

    def __init__(self, data: dict, state: ConnectionState):
        self.state = state
        self.from_dict = data

    @property
    def id(self) -> int:
        """Gives the Voice Channel' Snowflake ID

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    @property
    def guild(self) -> Guild:
        """The guild this channel is in

        Returns
        -------
        :class:`Guild`
        """
        id = self.from_dict["guild_id"]
        raw = self.state.app.fetch_raw_guild(guild_id=id)
        return Guild(raw, self.state.app.factory)

    @property
    def name(self) -> str:
        """The name of the voice channel

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["name"]

    @property
    def position(self) -> int:
        """The voice channel' position

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["position"]

    def permission_overwrites(self) -> list[str, Union[int, str]]:
        return self.from_dict["permission_overwrites"]


class DMChannel:
    """Represents a Discord DM Channel

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw data
    state: :class:`ConnectionState`
        The connection state
    """

    def __init__(self, data: dict, state: ConnectionState):
        self.from_dict = data
        self.state = state

    def last_message_id(self) -> int:
        """The snowflake id of the last message

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["last_message_id"]

    @property
    def id(self) -> int:
        """The snowflake id of the channel

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    def recipients(self):
        """The list of users in the channel

        Returns
        -------
        List[:class:`User`]
        """
        return [User(user_data) for user_data in self.from_dict["recipients"]]


def parse_groupdm_icon(format: FormatType, group_id: int, group_icon_hash: str) -> str:
    return f"https://cdn.discordapp.com/icons/{group_id}/{group_icon_hash}.{format}"


class GroupDMChannel(DMChannel):
    """Represents a Discord Group DM Channel

    .. versionadded:: 0.8.0

    Parameters
    -----------
    data: :class:`dict`
    """

    def name(self) -> str:
        """Gives the name of the Group DM

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["name"]

    def icon(self, format: FormatType = FormatType.PNG) -> str:
        """Gives the link of the channel' icon

        Returns
        -------
        :class:`str`
        """
        return parse_groupdm_icon(format, self.id, self.from_dict["icon"])

    def owner(self) -> User:
        """Returns the User which is the owner of this Group DM

        Returns
        -------
        :class:`User`
        """
        user = self.state.app.factory.get_user(self.from_dict["owner"])
        return User(user)


class Thread:
    """Represents a Discord Thread

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw thread data
    state: :class:`ConnectionState`
        The connection state
    """

    def __init__(self, data: dict, state: ConnectionState):
        self.from_dict = data
        self.state = state

    @property
    def id(self) -> int:
        """The thread' snowflake id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    @property
    def guild_id(self) -> int:
        """The guild id of the thread

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["guild_id"]

    @property
    def channel_id(self) -> int:
        """The channel id of the thread

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["parent_id"]

    @property
    def owner_id(self) -> int:
        """Gives the owner id of the Thread

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["owner_id"]

    @property
    def name(self) -> str:
        """Gives the name of the thread

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["name"]

    def last_message_id(self) -> int:
        """Gives the last message id in the thread

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["last_message_id"]

    def message_count(self) -> int:
        """Gives the amount of messages in the thread

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["message_count"]

    def member_count(self) -> int:
        """Gives the thread' member count

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["member_count"]

    @property
    def metadata(self) -> "ThreadMetadata":
        """Gives the thread' metadata

        Returns
        -------
        :class:`ThreadMetadata`
        """
        return ThreadMetadata(self.from_dict["thread_metadata"])


class ThreadMetadata:
    """Represents a Thread' metadata

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The metadata
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def archived(self) -> bool:
        """If the thread is archived

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["archived"]

    @property
    def auto_archive_duration(self) -> int:
        """The archived duration

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["auto_archive_duration"]

    @property
    def archive_timestamp(self) -> str:
        """The archived timestamp

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["archive_timestamp"]

    @property
    def locked(self) -> bool:
        """If the thread is locked or not

        Returns
        -------
        :class:`bool`
        """
        return self.from_dict["locked"]


class ThreadMember:
    """Represents a Discord Thread Member

    .. versionadded:: 0.8.0

    Parameters
    ----------
    data: :class:`dict`
        The raw Member data
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def id(self) -> int:
        """Gives the thread id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    @property
    def user_id(self) -> int:
        """Gives the thread members user id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["user_id"]

    @property
    def join_timestamp(self) -> str:
        """Gives the thread members join time

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["join_timestamp"]

    @property
    def flags(self) -> int:
        """Gives the thread members flags

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["flags"]


class StageInstance:
    """Represents a Discord Stage Instance

    .. versionadded:: 0.8.0

    Parameters
    -----------
    data: :class:`dict`
        The raw stage instance data
    """

    def __init__(self, data: dict):
        self.from_dict = data

    @property
    def id(self) -> int:
        """Gives the stage instance id

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["id"]

    @property
    def guild_id(self) -> int:
        """Gives the guild id of the stage instance

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["guild_id"]

    @property
    def channel_id(self) -> int:
        """Gives the channel id of the stage instance

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["channel_id"]

    @property
    def topic(self) -> str:
        """Gives the stage instance' topic

        Returns
        -------
        :class:`str`
        """
        return self.from_dict["topic"]

    def privacy_level(self) -> int:
        """Gives the stage instance' privacy level

        Returns
        -------
        :class:`int`
        """
        return self.from_dict["privacy_level"]



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE

from typing import List

__all__: List[str] = ["MessageFlags"]


class MessageFlags:
    """Represents a discord message flag object

    .. versionadded:: 0.7.0
    """

    @classmethod
    def CROSSPOSTED(self):
        return 1 << 0

    @classmethod
    def IS_CROSSPOSTED(self):
        return 1 << 1

    @classmethod
    def SUPPRESS_EMBEDS(self):
        return 1 << 2

    @classmethod
    def SOURCE_MESSAGE_DELETED(self):
        return 1 << 3

    @classmethod
    def URGENT(self):
        return 1 << 4

    @classmethod
    def HAS_THREAD(self):
        return 1 << 5

    @classmethod
    def EPHEMERAL(self):
        return 1 << 6

    @classmethod
    def LOADING(self):
        return 1 << 7



# -*- coding: utf-8 -*-
# cython: language_level=3
# Copyright (c) 2021-present VincentRPS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE


class Enum:
    def __new__(self, name, value):
        setattr(self, name, value)


class ButtonStyle(Enum):
    PRIMARY = 1
    SECONDARY = 2
    SUCCESS = 3
    DANGER = 4
    LINK = 5
    BLURPLE = 1
    GRAY = 2
    GREEN = 3
    RED = 4


class ChannelType(Enum):
    GUILD_TEXT = 0
    DM = 1
    GUILD_VOICE = 2
    GROUP_DM = 3
    GUILD_CATEGORY = 4
    GUILD_NEWS = 5
    GUILD_STORE = 6  # probably gonna be deprecated soon
    GUILD_NEWS_THREAD = 10
    GUILD_PUBLIC_THREAD = 11
    GUILD_PRIVATE_THREAD = 12
    GUILD_STAGE_VOICE = 13


class VideoQuality(Enum):
    AUTO = 1
    FULL = 2


class ApplicationCommandType(Enum):
    CHAT_INPUT = 1
    USER = 2
    MESSAGE = 3


class StickerType(Enum):
    STANDARD = 1
    GUILD = 2


class StickerFormatType(Enum):
    PNG = 1
    APNG = 2
    LOTTIE = 3


class FormatType(Enum):
    JPEG = ".jpeg"
    PNG = ".png"
    WEBP = ".webp"
    GIF = ".gif"
    LOTTIE = ".json"


class ScheduledEventStatusType(Enum):
    SCHEDULED = 1
    ACTIVE = 2
    COMPLETED = 3
    CANCELED = 4


class ScheduledEventType(Enum):
    STAGE_INSTANCE = 1
    VOICE = 2
    EXTERNAL = 3


class ModalStyle(Enum):
    SHORT = 1
    PARAGRAPH = 2
